#include "Blackboard.H"
#include "ChargeMap.H"
#include "TBConfig.H"
#include "UnitAgent.H"

using namespace std;

vector<BlackboardItem> Blackboard::items;

const sint4 Blackboard::AC_NO_ACTION = 1;
const sint4 Blackboard::AC_MOVE = 2;
const sint4 Blackboard::AC_ATTACK_BASE = 3;
const sint4 Blackboard::AC_ATTACK_TANK = 4;
const sint4 Blackboard::AC_ATTACK_MARINE = 5;
const sint4 Blackboard::AC_AWAIT_COOLDOWN = 6;

const sint4 Blackboard::TYPE_TANK = 0;
const sint4 Blackboard::TYPE_MARINE = 1;
const sint4 Blackboard::TYPE_BASE = 2;

const sint4 Blackboard::NO_ID = -1;

sint4 Blackboard::state = 1;
const sint4 Blackboard::STATE_HUNT_UNITS = 1;
const sint4 Blackboard::STATE_HUNT_BASES = 2;

/*
	Switch state for the overall tactics.
*/
void Blackboard::setState(sint4 mState) {
	state = mState;
}

/*
	Updates the tactics state based on the current game state.
*/
void Blackboard::updateState(const Game& game) {
	int n = game.get_player_num();
	const sint4 cid = game.get_client_player();
	
	int friendlyUnits = 0;
	int friendlyBases = 0;
	int enemyUnits = 0;
	int enemyBases = 0;

	//Iterate through all players.
	for (int player = 0; player < n; player++) {
		const Game::ObjCont &objs = game.get_objs(player);

		bool friendly = false;
		if (player == cid) {
			friendly = true;
		}

		//Iterate through all objects for this player
		FORALL (objs, it) {
			GameObjPtr gob = (*it)->get_GameObj();
			if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
				//Valid object found
				std::string objType = gob->bp_name();
				
				if (objType == "tank" || objType == "marine") {
					if (friendly) friendlyUnits++;
					else enemyUnits++;
				}
				if (objType == "controlCenter") {
					if (friendly) friendlyBases++;
					else enemyBases++;
				}
			}
		}
	}

	sint4 newState = Blackboard::STATE_HUNT_UNITS;

	if (enemyUnits <= 3) {
		newState = Blackboard::STATE_HUNT_BASES;
	}
	else {
		int unitRatio = friendlyUnits / enemyUnits;
		if (unitRatio >= TBConfig::params.Blackboard_HuntUnits_Ratio) newState = Blackboard::STATE_HUNT_BASES;
	}

	if (friendlyBases == 1 && enemyUnits > 0) {
		newState = Blackboard::STATE_HUNT_UNITS;
	}
	if (enemyBases == 1) {
		newState = Blackboard::STATE_HUNT_BASES;
	}

	if (friendlyBases == 0) {
		//This is probably a game without bases
		newState = Blackboard::STATE_HUNT_UNITS;
	}
	
	if (newState != Blackboard::state) {
		Blackboard::state = newState;
	}	
}

/*
	Gets the type (tank, marine or base) for a unit.
*/
int Blackboard::getType(GameObjPtr unit) {
	std::string objType = unit->bp_name();

	if (objType == "tank") {
		return Blackboard::TYPE_TANK;
	}
	else if (objType == "marine") {
		return Blackboard::TYPE_MARINE;
	}
	else if (objType == "controlCenter") {
		return Blackboard::TYPE_BASE;
	}
	else {
		cout << "Unknown unit type:" << objType << endl;
		return Blackboard::TYPE_BASE;
	}
}

/*
	Sets the current action for a player unit. If targetX, targetY or targetId is unknown, use the NO_ID
	constant.
*/
void Blackboard::setAction(int objX, int objY, uint4 objId, int objType, int action, int targetX, int targetY, uint4 targetId, bool offensive) {
	//First, iterate to see if this object already has an entry
	vector<BlackboardItem>::iterator it = Blackboard::items.begin();
	while (it != Blackboard::items.end()) {
		if ((*it).isThisObj(objId)) {
			//Entry found. Update values
			(*it).objX = objX;
			(*it).objY = objY;
			(*it).actionId = action;
			(*it).targetX = targetX;
			(*it).targetY = targetY;
			(*it).targetId = targetId;
			(*it).offensive = offensive;

			if (action != Blackboard::AC_NO_ACTION) {
				//Action performed.
				if ((*it).isRestless()) {
					//If unit is restless, it has do do some random actions before
					//the idle counter is reset
					if ((*it).restlessMoveCnt > TBConfig::params.Blackboard_ActionsBeforeClearRestless) {
						//Enough random actions. Reset.
						(*it).restlessMoveCnt = 0;
						(*it).idleCnt = 0;
						(*it).restlessDirection = 0;
					}
					else {
						//Increase random actions
						(*it).restlessMoveCnt = (*it).restlessMoveCnt + 1;
					}
				}
				else {
					//Not restless. Reset counter.
					(*it).restlessMoveCnt = 0;
					(*it).idleCnt = 0;
					(*it).restlessDirection = 0;
				}
			}
			else {
				//No action. Increase idle counter.
				(*it).idleCnt = (*it).idleCnt + 1;
			}

			//Done. Return
			return;
		}

		it++;
	}

	//If no entry was found, we need to add it
	BlackboardItem item;
	item.type = objType;
	item.objX = objX;
	item.objY = objY;
	item.objId = objId;
	item.actionId = action;
	item.targetX = targetX;
	item.targetY = targetY;
	item.targetId = targetId;
	item.offensive = offensive;
	item.idleCnt = 0;
	item.restlessMoveCnt = 0;
	item.restlessDirection = 0;
	item.alive = true;

	Blackboard::items.push_back(item);
}

/*
	Adds a blackboard item for a friendly base. Is necessary to keep track of the location of the bases.
*/
void Blackboard::addBase(int objX, int objY, uint4 objId) {
	vector<BlackboardItem>::iterator it = Blackboard::items.begin();
	while (it != Blackboard::items.end()) {
		if ((*it).isBase() && (*it).isThisObj(objId)) {
			//Base already in list
			return;
		}
		it++;
	}

	BlackboardItem item;
	item.type = Blackboard::TYPE_BASE;
	item.objX = objX;
	item.objY = objY;
	item.objId = objId;
	item.actionId = Blackboard::NO_ID;
	item.targetX = Blackboard::NO_ID;
	item.targetY = Blackboard::NO_ID;
	item.targetId = Blackboard::NO_ID;
	item.offensive = false;
	item.idleCnt = 0;
	item.alive = true;

	Blackboard::items.push_back(item);
}

/*
	Each time a tank has been destroyed, we need to update the alive status for the
	elements in the blackboard.
*/
void Blackboard::updateAliveStatus(const Game& game, sint4 playerId) {
	PlayerInfo& pi = game.get_cplayer_info();

	vector<BlackboardItem>::iterator bl_it = Blackboard::items.begin();
	while (bl_it != Blackboard::items.end()) {
		bool found = false;
		
		const Game::ObjCont &objs = game.get_objs(playerId);
		FORALL (objs, it) {
			GameObj *gob = (*it)->get_GameObj();
			if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
				uint4 objId = pi.get_id(gob);
				if ((*bl_it).isThisObj(objId)) {
					found = true;
					break;
				}
			}
		}

		if (!found) {
			(*bl_it).alive = false;
		}

		bl_it++;
	}
}

/*
	Checks if a unit is lonely. A unit is considered lonely if there are less
	than four nearby firendly units.
*/
bool Blackboard::isLonely(int unitX, int unitY, uint4 objId) {
	int cnt = 0;

	vector<BlackboardItem>::iterator it = Blackboard::items.begin();
	while (it != Blackboard::items.end()) {
		if (!(*it).isThisObj(objId) && (*it).isAlive() && (*it).isUnit()) {
			//Other object found. Check distance
			real8 dist = Blackboard::distance(unitX, unitY, (*it).objX, (*it).objY);
			if (dist <= TBConfig::params.Blackboard_MaxDistBetweenGroupUnits) {
				//We found another object within max_dist. 
				cnt++;
			}
		}

		it++;
	}

	//Less than x units nearby. This unit is considered lonely 
	if (cnt <= TBConfig::params.Blackboard_MinUnitsInGroup - 1) return true;
	
	return false;
}

/*
	Returns true if a unit is restless, i.e. has not moved or shoot for
	several frames.
*/
bool Blackboard::isRestless(uint4 objId) {
	vector<BlackboardItem>::iterator it = Blackboard::items.begin();
	while (it != Blackboard::items.end()) {
		if ((*it).isThisObj(objId) && (*it).isAlive() && (*it).isUnit() && (*it).isRestless()) {
			//Object found and is restless
			return true;
		}

		it++;
	}

	return false;
}

/*
	Gets the directon used in restless move for this unit. If 0, no direction has been set.
*/
int Blackboard::getDirection(uint4 objId) {
	vector<BlackboardItem>::iterator it = Blackboard::items.begin();
	while (it != Blackboard::items.end()) {
		if ((*it).isThisObj(objId) && (*it).isAlive() && (*it).isUnit() && (*it).isRestless()) {
			//Object found and is restless. Return the direction
			return (*it).restlessDirection;
		}

		it++;
	}

	return 0;
}

/*
	Sets the directon used in restless move for this unit. Direction can be between 1 and 8.
*/
void Blackboard::setDirection(uint4 objId, int direction) {
	vector<BlackboardItem>::iterator it = Blackboard::items.begin();
	while (it != Blackboard::items.end()) {
		if ((*it).isThisObj(objId) && (*it).isAlive() && (*it).isUnit() && (*it).isRestless()) {
			(*it).restlessDirection = direction;
		}

		it++;
	}
}

/*
	Calculates the distance between two positions.
*/
real8 Blackboard::distance(int x1, int y1, int x2, int y2) {
	real8 ASquare = (x2 - x1) * (x2 - x1);
	real8 BSquare = (y2 - y1) * (y2 - y1);

	real8 dist = sqrt(ASquare + BSquare);
	return dist;
}

/*
	Gets the attack multiplier for attacking an enemy tank. The attack value increases
	the more player units attacking the target, but reduces to 0 when enough units
	attacks the target to reduce it's HP to 0.
*/
real8 Blackboard::getTankAttackMultiplier(GameObjPtr target, uint4 targetId) {	
	int currentHP = target->get_int("hp");
	return getAttackMultiplier(UnitAgent::TANK_MIN_DMG, UnitAgent::TANK_MAX_DMG, targetId, currentHP, UnitAgent::TANK_HP, UnitAgent::TANK_ARMOR);
}

/*
	Gets the attack multiplier for attacking an enemy marine. The attack value increases
	the more player units attacking the target, but reduces to 0 when enough units
	attacks the target to reduce it's HP to 0.
*/
real8 Blackboard::getMarineAttackMultiplier(GameObjPtr target, uint4 targetId) {
	int currentHP = target->get_int("hp");
	return getAttackMultiplier(UnitAgent::MARINE_MIN_DMG, UnitAgent::MARINE_MAX_DMG, targetId, currentHP, UnitAgent::MARINE_HP, UnitAgent::MARINE_ARMOR);
}

/*
	Gets the attack multiplier for attacking an enemy base. The attack value increases
	the more player units attacking the target, but reduces to 0 when enough units
	attacks the target to reduce it's HP to 0.
*/
real8 Blackboard::getBaseAttackMultiplier(GameObjPtr target, uint4 targetId) {
	int currentHP = target->get_int("hp");
	//NOTE: Bases are only present in games without marines, i.e. a base is always attacked by a tank.
	return getAttackMultiplier(UnitAgent::TANK_MIN_DMG, UnitAgent::TANK_MAX_DMG, targetId, currentHP, UnitAgent::CONTROLCENTER_HP, UnitAgent::CONTROLCENTER_ARMOR);
}

/*
	Gets the attack multiplier for attacking an enemy unit or base. The attack value increases
	the more player units attacking the target, but reduces to 0 when enough units
	attacks the target to reduce it's HP to 0.
*/
real8 Blackboard::getAttackMultiplier(sint4 unitMinDmg, sint4 unitMaxDmg, uint4 targetId, int targetCurrentHP, int targetMaxHP, int targetArmor) {
	//Calculate the mean damage each attack does
	real8 range = (real8)unitMaxDmg - (real8)unitMinDmg;
	real8 useDmg = (real8)unitMinDmg + TBConfig::params.Attack_DmgProbability * range - (real8)targetArmor;

	//Calculate number of attacks needed to destroy the unit (based on current HP)
	real8 neededAttacks = ceil((real8)targetCurrentHP / useDmg);
	//Calculate number of attacks needed to destroy the unit (based on max HP)
	real8 maxAttacks = ceil((real8)targetMaxHP / useDmg);

	//Get number of units currently attacking the target
	real8 noAttacks = (real8)calculateAttacksAtTarget(targetId);
	
	real8 returnVal = 0;
	if (noAttacks <= maxAttacks) {
		//More attacks needed on the target
		returnVal = (real8)(maxAttacks - neededAttacks + noAttacks + 1);
	}
	else {
		//Enough attacks already on target
		returnVal = 0;
	}
	
	return returnVal;
}

/*
	Calculates the number of units attacking a target the current action frame.
*/
int Blackboard::calculateAttacksAtTarget(uint4 targetId) {
	vector<BlackboardItem>::iterator it = Blackboard::items.begin();
	
	int cnt = 0;
	while (it != Blackboard::items.end()) {
		if ((*it).isAlive() && (*it).attacksTarget(targetId)) {
			cnt++;
		}

		it++;
	}

	return cnt;
}

/*
	Prints a debug output with the current actions for the player units.
*/
void Blackboard::printDebug() {
	cout << "" << endl;
	cout << "---- Blackboard ----" << endl;

	cout << "Current tactics: ";
	if (Blackboard::state == Blackboard::STATE_HUNT_UNITS) {
		cout << "Hunt enemy units";
	}
	if (Blackboard::state == Blackboard::STATE_HUNT_BASES) {
		cout << "Hunt enemy bases";
	}
	cout << "" << endl;

	cout << "--------------------" << endl;

	vector<BlackboardItem>::iterator it = Blackboard::items.begin();
	while (it != Blackboard::items.end()) {
		cout << "[" << (*it).objX << "," << (*it).objY << "] " << (*it).objId << ".";

		if ((*it).isTank()) {
			cout << "Tank ";
		}
		if ((*it).isMarine()) {
			cout << "Marine ";
		}
		if ((*it).isBase()) {
			cout << "Base ";
		}

		if ((*it).isAlive() && (*it).isUnit()) {
			cout << (*it).actionId << ".";
			if ((*it).actionId == AC_NO_ACTION) cout << "no_action ";
			if ((*it).actionId == AC_MOVE) cout << "move ";
			if ((*it).actionId == AC_ATTACK_BASE) cout << "attack_base ";
			if ((*it).actionId == AC_ATTACK_TANK) cout << "attack_tank ";
			if ((*it).actionId == AC_ATTACK_MARINE) cout << "attack_marine ";
			if ((*it).actionId == AC_AWAIT_COOLDOWN) cout << "await_cooldown ";

			if ((*it).targetId != Blackboard::NO_ID) {
				cout << (*it).targetId << " ";
			}
			if ((*it).targetX != Blackboard::NO_ID) {
				cout << "[" << (*it).targetX << "," << (*it).targetY << "] ";
			}

			if ((*it).isRestless()) {
				cout << "restless ";
			}

			if ((*it).offensive) cout << "offensive" << endl;
			else cout << "defensive" << endl;
		}
		if ((*it).isAlive() && (*it).isBase()) {
			cout << "" << endl;
		}
		else if (!(*it).isAlive()) {
			cout << "DESTROYED" << endl;
		}

		it++;
	}

	cout << "--------------------" << endl;
	cout << "" << endl;
}

/*
	Prints a debug line to the console containing current
	number of friendly or enemy units and bases.

	This is used by the GA optimization.
*/
void Blackboard::printUnitStatus(const Game& game) {
	int n = game.get_player_num();
	const sint4 cid = game.get_client_player();
	
	int friendlyUnits = 0;
	int friendlyBases = 0;
	int enemyUnits = 0;
	int enemyBases = 0;

	//Iterate through all players.
	for (int player = 0; player < n; player++) {
		const Game::ObjCont &objs = game.get_objs(player);

		bool friendly = false;
		if (player == cid) {
			friendly = true;
		}

		//Iterate through all objects for this player
		FORALL (objs, it) {
			GameObjPtr gob = (*it)->get_GameObj();
			if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
				//Valid object found
				std::string objType = gob->bp_name();
				
				if (objType == "tank" || objType == "marine") {
					if (friendly) friendlyUnits++;
					else enemyUnits++;
				}
				if (objType == "controlCenter") {
					if (friendly) friendlyBases++;
					else enemyBases++;
				}
			}
		}
	}

	cout << "STATUS cid " << cid << " FriendlyBases " << friendlyBases << " EnemyBases " << enemyBases << " FriendlyUnits " << friendlyUnits << " EnemyUnits " << enemyUnits << endl;
}
