#ifndef Blackboard_H
#define Blackboard_H

#include "GameObj.H"
#include "Game.H"
#include "TBConfig.H"

using namespace std;

/*
	Struct for blackboard items.
*/
struct BlackboardItem {
	uint4 type;

	sint4 objX;
	sint4 objY;
	sint4 objId;

	sint4 idleCnt;
	sint4 restlessMoveCnt;
	sint4 restlessDirection;
	
	sint4 targetX;
	sint4 targetY;
	sint4 targetId;
	
	sint4 actionId;
	bool offensive;
	bool alive;

	bool isAlive() {
		return alive;
	}

	bool isRestless() {
		if (idleCnt >= TBConfig::params.Blackboard_IdleFramesBeforeRestless) {
			return true;
		}

		return false;
	}

	bool hasDirection() {
		if (restlessDirection == 0) return false;
		else return true;
	}

	bool isThisObj(sint4 mObjId) {
		if (objId == -1) return false; //Blackboard::NO_ID
		if (objId == mObjId) return true;
		
		return false;
	}

	bool attacksTarget(sint4 mTargetId) {
		if (targetId == -1) return false; //Blackboard::NO_ID

		if (targetId == mTargetId) {
			if (actionId == 3) return true; //Blackboard::AC_ATTACK_BASE
			if (actionId == 4) return true; //Blackboard::AC_ATTACK_TANK
			if (actionId == 5) return true; //Blackboard::AC_ATTACK_MARINE
		}

		return false;
	}

	bool isUnit() {
		if (type == 0) return true;
		if (type == 1) return true;
		return false;
	}

	bool isTank() {
		if (type == 0) return true;
		return false;
	}

	bool isMarine() {
		if (type == 1) return true;
		return false;
	}

	bool isBase() {
		if (type == 2) return true;
		return false;
	}
};

/*
	This class represents a blackboard where the player units can communicate there actions to
	other player units. It also contains some methods for unit action coordination.
*/
class Blackboard {
	
	typedef GameObj* GameObjPtr;

	public:

		/*
			Gets the type (tank, marine or base) for a unit.
		*/
		static int getType(GameObjPtr unit);

		/*
			Sets the current action for a player unit. If targetX, targetY or targetId is unknown, use the NO_ID
			constant.
		*/
		static void setAction(int objX, int objY, uint4 objId, int objType, int action, int targetX, int targetY, uint4 targetId, bool offensive);

		/*
			Adds a blackboard item for a friendly base. Is necessary to keep track of the location of the bases.
		*/
		static void addBase(int objX, int objY, uint4 objId);

		/*
			Each time a tank has been destroyed, we need to update the alive status for the
			elements in the blackboard.
		*/
		static void updateAliveStatus(const Game& game, sint4 playerId);

		/*
			Checks if a unit is lonely. A unit is considered lonely if there are less
			than four nearby firendly units.
		*/
		static bool isLonely(int unitX, int unitY, uint4 objId);

		/*
			Returns true if a unit is restless, i.e. has not moved or shoot for
			several frames.
		*/
		static bool isRestless(uint4 objId);

		/*
			Gets the directon used in restless move for this unit. If 0, no direction has been set.
		*/
		static int getDirection(uint4 objId);

		/*
			Sets the directon used in restless move for this unit. Direction can be between 1 and 8.
		*/
		static void setDirection(uint4 objId, int direction);
		
		/*
			Calculates the distance between two positions.
		*/
		static real8 distance(int x1, int y1, int x2, int y2);

		/*
			Gets the attack multiplier for attacking an enemy tank. The attack value increases
			the more player units attacking the target, but reduces to 0 when enough units
			attacks the target to reduce it's HP to 0.
		*/
		static real8 getTankAttackMultiplier(GameObjPtr target, uint4 targetId);

		/*
			Gets the attack multiplier for attacking an enemy marine. The attack value increases
			the more player units attacking the target, but reduces to 0 when enough units
			attacks the target to reduce it's HP to 0.
		*/
		static real8 getMarineAttackMultiplier(GameObjPtr target, uint4 targetId);

		/*
			Gets the attack multiplier for attacking an enemy base. The attack value increases
			the more player units attacking the target, but reduces to 0 when enough units
			attacks the target to reduce it's HP to 0.
		*/
		static real8 getBaseAttackMultiplier(GameObjPtr target, uint4 targetId);

		/*
			Gets the attack multiplier for attacking an enemy unit or base. The attack value increases
			the more player units attacking the target, but reduces to 0 when enough units
			attacks the target to reduce it's HP to 0.
		*/
		static real8 getAttackMultiplier(sint4 unitMinDmg, sint4 unitMaxDmg, uint4 targetId, int targetCurrentHP, int targetMaxHP, int targetArmor);

		/*
			Switch state for the overall tactics.
		*/
		static void setState(sint4 mState);

		/*
			Updates the tactics state based on the current game state.
		*/
		static void updateState(const Game& game);

		//The unit is idle
		static const sint4 AC_NO_ACTION;
		//The unit is moving
		static const sint4 AC_MOVE;
		//The unit is attacking a base
		static const sint4 AC_ATTACK_BASE;
		//The unit is attacking a tank
		static const sint4 AC_ATTACK_TANK;
		//The unit is attacking a marine
		static const sint4 AC_ATTACK_MARINE;
		//The unit is awaiting cooldown
		static const sint4 AC_AWAIT_COOLDOWN;

		//Blackboard item for bases
		static const sint4 TYPE_BASE;
		//Blackboard item for tanks
		static const sint4 TYPE_TANK;
		//Blackboard item for marines
		static const sint4 TYPE_MARINE;

		//Default value to use if no value (location/id) exists
		static const sint4 NO_ID;

		//Current state for general tactics
		static sint4 state;
		//Prioritize hunt enemy units
		static const sint4 STATE_HUNT_UNITS;
		//Prioritize hunt enemy bases
		static const sint4 STATE_HUNT_BASES;

		/*
			Prints a debug output with the current actions for the player units.
		*/
		static void printDebug();

		/*
			Prints a debug line to the console containing current
			number of friendly or enemy units and bases.

			This is used by the GA optimization.
		*/
		static void printUnitStatus(const Game& game);

	private:

		/*
			Calculates the number of units attacking a target the current action frame.
		*/
		static int calculateAttacksAtTarget(uint4 targetId);
		
		//Contains the blackboard action items
		static vector<BlackboardItem> items;

};

#endif
