#include "ChargeMap.H"
#include "Game.H"
#include "Blackboard.H"

using namespace std;

const int ChargeMap::CMTILE_WIDTH = 8;
const int ChargeMap::CHARGE_IMPASSABLE_TILES = -1000;

/*
	Constructor.
*/
ChargeMap::ChargeMap(GameStateModulePtr mState) {
	int tile_w = mState->get_game().get_tile_points();

	width = mState->get_game().get_map().get_width() * tile_w / CMTILE_WIDTH;
	height = mState->get_game().get_map().get_height() * tile_w / CMTILE_WIDTH;

	for (int cX = 0; cX < width; cX++) {
		for (int cY = 0; cY < height; cY++) {
			map[cX][cY].enemyBaseCharge_huntUnits = 0;
			map[cX][cY].enemyBaseCharge_huntBases = 0;
			map[cX][cY].enemyUnitOffensiveCharge = 0;
			map[cX][cY].enemyUnitDefensiveCharge = 0;

			map[cX][cY].playerBaseCharge = 0;
			map[cX][cY].playerUnitCharge = 0;

			map[cX][cY].neutralUnitCharge = 0;

			map[cX][cY].terrainCharge = 0;
		}
	}
}

/*
	Destructor.
*/
ChargeMap::~ChargeMap() {
	
}

/*
	Clears the unit charge map. Shall be called before
	updating the unit charges each frame.
*/
void ChargeMap::clearUnitCharge() {
	for (int cX = 0; cX < width; cX++) {
		for (int cY = 0; cY < height; cY++) {
			map[cX][cY].enemyUnitOffensiveCharge = 0;
			map[cX][cY].enemyUnitDefensiveCharge = 0;
			map[cX][cY].playerUnitCharge = 0;
			map[cX][cY].neutralUnitCharge = 0;
		}
	}
}

/*
	Clears the base charge map. Shall be called before
	updating the base charges each time a friendly or
	enemy base is destroyed.
*/
void ChargeMap::clearBaseCharge() {
	for (int cX = 0; cX < width; cX++) {
		for (int cY = 0; cY < height; cY++) {
			map[cX][cY].enemyBaseCharge_huntUnits = 0;
			map[cX][cY].enemyBaseCharge_huntBases = 0;
			map[cX][cY].playerBaseCharge = 0;
		}
	}
}

/*
	Clears the unit charge in a specific point.
*/
void ChargeMap::clearUnitCharge(sint4 x, sint4 y) {
	map[x][y].enemyUnitOffensiveCharge = 0;
	map[x][y].enemyUnitDefensiveCharge = 0;
	map[x][y].playerUnitCharge = 0;
	map[x][y].neutralUnitCharge = 0;
}

/*
	Updates the charge an enemy base has in a specific point.
*/
void ChargeMap::updateEnemyBaseCharge_huntUnits(sint4 x, sint4 y, real8 charge) {
	if (x < 0 || y < 0) return;
	if (x >= width || y >= height) return;

	map[x][y].enemyBaseCharge_huntUnits += charge;
}

/*
	Updates the charge an enemy base has in a specific point.
*/
void ChargeMap::updateEnemyBaseCharge_huntBases(sint4 x, sint4 y, real8 charge) {
	if (x < 0 || y < 0) return;
	if (x >= width || y >= height) return;

	map[x][y].enemyBaseCharge_huntBases += charge;
}

/*
	Updates the charge an enemy tank has in a specific point during offensive mode.
*/
void ChargeMap::updateEnemyUnitOffensiveCharge(sint4 x, sint4 y, real8 charge) {
	if (x < 0 || y < 0) return;
	if (x >= width || y >= height) return;

	map[x][y].enemyUnitOffensiveCharge += charge;
}

/*
	Updates the charge an enemy tank has in a specific point during defensive mode.
*/
void ChargeMap::updateEnemyUnitDefensiveCharge(sint4 x, sint4 y, real8 charge) {
	if (x < 0 || y < 0) return;
	if (x >= width || y >= height) return;

	map[x][y].enemyUnitDefensiveCharge += charge;
}

/*
	Updates the charge a player base has in a specific point.
*/
void ChargeMap::updatePlayerBaseCharge(sint4 x, sint4 y, real8 charge) {
	if (x < 0 || y < 0) return;
	if (x >= width || y >= height) return;

	map[x][y].playerBaseCharge += charge;
}

/*
	Updates the charge a player unit has in a specific point.
*/
void ChargeMap::updatePlayerUnitCharge(sint4 x, sint4 y, real8 charge) {
	if (x < 0 || y < 0) return;
	if (x >= width || y >= height) return;

	map[x][y].playerUnitCharge += charge;
}

/*
	Updates the charge a neutral unit has in a specific point.
*/
void ChargeMap::updateNeutralUnitCharge(sint4 x, sint4 y, real8 charge) {
	if (x < 0 || y < 0) return;
	if (x >= width || y >= height) return;

	map[x][y].neutralUnitCharge += charge;
}

/*
	Updates the charge a terrain tile has in a specific point.
*/
void ChargeMap::updateTerrainCharge(sint4 x, sint4 y, real8 charge) {
	if (x < 0 || y < 0) return;
	if (x >= width || y >= height) return;

	if (charge < map[x][y].terrainCharge) {
		map[x][y].terrainCharge = charge;
	}
}

/*
	Sets the charge a terrain tile has in a specific point.
	Previous charge in the point is discarded.
*/
void ChargeMap::setTerrainCharge(sint4 x, sint4 y, real8 charge) {
	if (x < 0 || y < 0) return;
	if (x >= width || y >= height) return;

	map[x][y].terrainCharge = charge;
}

/*
	Returns the charge in a specific point.
*/
real8 ChargeMap::getCharge(sint4 x, sint4 y, bool offensive) {
	if (offensive) return getOffensiveCharge(x, y);
	else return getDefensiveCharge(x, y);
}

bool ChargeMap::canSpotEnemyUnit(sint4 x, sint4 y, bool offensive) {
	if (offensive) {
		if (map[x][y].enemyUnitOffensiveCharge == 0) {
			return false;
		}
	}
	else {
		if (map[x][y].enemyUnitDefensiveCharge == 0) {
			return false;
		}
	}

	return true;
}
/*
	Returns the charge in a specific point for units in offensive mode.
*/
real8 ChargeMap::getOffensiveCharge(sint4 x, sint4 y) {
	real8 baseCharge = 0;
	if (Blackboard::state == Blackboard::STATE_HUNT_BASES) {
		baseCharge = map[x][y].enemyBaseCharge_huntBases;
	}
	if (Blackboard::state == Blackboard::STATE_HUNT_UNITS) {
		baseCharge = map[x][y].enemyBaseCharge_huntUnits + 0.1 * map[x][y].enemyBaseCharge_huntBases;
	}

	return map[x][y].terrainCharge + map[x][y].neutralUnitCharge + baseCharge + map[x][y].playerBaseCharge + map[x][y].playerUnitCharge + map[x][y].enemyUnitOffensiveCharge;
}

/*
	Returns the charge in a specific point for units in defensive mode.
*/
real8 ChargeMap::getDefensiveCharge(sint4 x, sint4 y) {
	real8 baseCharge = 0;
	if (Blackboard::state == Blackboard::STATE_HUNT_BASES) {
		baseCharge = map[x][y].enemyBaseCharge_huntBases;
	}
	if (Blackboard::state == Blackboard::STATE_HUNT_UNITS) {
		baseCharge = map[x][y].enemyBaseCharge_huntUnits + 0.1 * map[x][y].enemyBaseCharge_huntBases;
	}

	return map[x][y].terrainCharge + map[x][y].neutralUnitCharge + baseCharge + map[x][y].playerBaseCharge + map[x][y].playerUnitCharge + map[x][y].enemyUnitDefensiveCharge;
}

/*
	Returns the enemy unit charge in a specific point
	for units in offensive mode.
	This method is only used in debug purpose.
*/
real8 ChargeMap::getUnitOffensiveCharge(sint4 x, sint4 y) {
	return map[x][y].terrainCharge + map[x][y].neutralUnitCharge + map[x][y].playerBaseCharge + map[x][y].playerUnitCharge + map[x][y].enemyUnitOffensiveCharge;
}

/*
	Returns the enemy unit charge in a specific point
	for units in defensive mode.
	This method is only used in debug purpose.
*/
real8 ChargeMap::getUnitDefensiveCharge(sint4 x, sint4 y) {
	return map[x][y].terrainCharge + map[x][y].neutralUnitCharge + map[x][y].playerBaseCharge + map[x][y].playerUnitCharge + map[x][y].enemyUnitDefensiveCharge;
}

/*
	Returns the enemy base charge in a specific point.
	This method is only used in debug purpose.
*/
real8 ChargeMap::getEnemyBaseCharge(sint4 x, sint4 y) {
	real8 baseCharge = 0;
	if (Blackboard::state == Blackboard::STATE_HUNT_BASES) {
		baseCharge = map[x][y].enemyBaseCharge_huntBases;
	}
	if (Blackboard::state == Blackboard::STATE_HUNT_UNITS) {
		baseCharge = map[x][y].enemyBaseCharge_huntUnits;
	}

	return map[x][y].terrainCharge + baseCharge;
}

/*
	Returns the player base and units charge in a specific point.
	This method is only used in debug purpose.
*/
real8 ChargeMap::getPlayerCharge(sint4 x, sint4 y) {
	return map[x][y].terrainCharge + map[x][y].neutralUnitCharge + map[x][y].playerBaseCharge + map[x][y].playerUnitCharge;
}

/*
	Returns the terrain charge in a specific point.
	This method is only used in debug purpose.
*/
real8 ChargeMap::getTerrainCharge(sint4 x, sint4 y) {
	if (x < 0 || x >= width - 1) return 0;
	if (y < 0 || y >= height - 1) return 0;

	return map[x][y].terrainCharge;
}

/*
	Returns minimal terrain charge in a specific point.
	If a tile is impassable due to terrain or a player base
	a negative charge is returned, otherwise no charge.
*/
real8 ChargeMap::getMinimalTerrainCharge(sint4 x, sint4 y) {
	if (map[x][y].terrainCharge <= ChargeMap::CHARGE_IMPASSABLE_TILES / 2) {
		return ChargeMap::CHARGE_IMPASSABLE_TILES;
	}
	else {
		//Check if we are inside any player bases
		if (map[x][y].playerBaseCharge <= -80) {
			//Inside
			return ChargeMap::CHARGE_IMPASSABLE_TILES;
		}
		else {
			//Outside. Return no charge
			return 0;
		}
	}
}

