#ifndef ChargeMap_H
#define ChargeMap_H

#include "GameStateModule.H"
#include "GameObj.H"
#include "Map.H"
using namespace std;

/*
	This class holds the charge map and contain methods for updating the charges and
	returning charge values in a specific tile.
*/
class ChargeMap {

	typedef GameStateModule* GameStateModulePtr;

	public:

		/*
			Constructor.
		*/
		ChargeMap(GameStateModulePtr mState);

		/*
			Destructor.
		*/
		~ChargeMap();

		/*
			Clears the unit charge map. Shall be called before
			updating the unit charges each frame.
		*/
		void clearUnitCharge();

		/*
			Clears the base charge map. Shall be called before
			updating the base charges each time a friendly or
			enemy base is destroyed.
		*/
		void clearBaseCharge();

		/*
			Clears the unit charge in a specific point.
		*/
		void clearUnitCharge(sint4 x, sint4 y);

		/*
			Updates the charge an enemy tank has in a specific point during offensive mode.
		*/
		void updateEnemyUnitOffensiveCharge(sint4 x, sint4 y, real8 charge);

		/*
			Updates the charge an enemy tank has in a specific point during defensive mode.
		*/
		void updateEnemyUnitDefensiveCharge(sint4 x, sint4 y, real8 charge);

		/*
			Updates the charge an enemy base has in a specific point.
		*/
		void updateEnemyBaseCharge_huntUnits(sint4 x, sint4 y, real8 charge);

		/*
			Updates the charge an enemy base has in a specific point.
		*/
		void updateEnemyBaseCharge_huntBases(sint4 x, sint4 y, real8 charge);

		/*
			Updates the charge a player unit has in a specific point.
		*/
		void updatePlayerUnitCharge(sint4 x, sint4 y, real8 charge);

		/*
			Updates the charge a player base has in a specific point.
		*/
		void updatePlayerBaseCharge(sint4 x, sint4 y, real8 charge);

		/*
			Updates the charge a neutral unit has in a specific point.
		*/
		void updateNeutralUnitCharge(sint4 x, sint4 y, real8 charge);

		/*
			Updates the charge a terrain tile has in a specific point.
		*/
		void updateTerrainCharge(sint4 x, sint4 y, real8 charge);

		/*
			Sets the charge a terrain tile has in a specific point.
			Previous charge in the point is discarded.
		*/
		void setTerrainCharge(sint4 x, sint4 y, real8 charge);

		/*
			Returns the charge in a specific point.
		*/
		real8 getCharge(sint4 x, sint4 y, bool offensive);

		bool canSpotEnemyUnit(sint4 x, sint4 y, bool offensive);

		/*
			Returns the charge in a specific point for units in offensive mode.
		*/
		real8 getOffensiveCharge(sint4 x, sint4 y);

		/*
			Returns the charge in a specific point for units in defensive mode.
		*/
		real8 getDefensiveCharge(sint4 x, sint4 y);

		/*
			Returns the terrain charge in a specific point.
			This method is only used in debug purpose.
		*/
		real8 getTerrainCharge(sint4 x, sint4 y);

		/*
			Returns minimal terrain charge in a specific point.
			If a tile is impassable due to terrain or a player base
			a negative charge is returned, otherwise no charge.
		*/
		real8 getMinimalTerrainCharge(sint4 x, sint4 y);

		/*
			Returns the enemy base charge in a specific point.
			This method is only used in debug purpose.
		*/
		real8 getEnemyBaseCharge(sint4 x, sint4 y);
		
		/*
			Returns the enemy unit charge in a specific point
			for units in offensive mode.
			This method is only used in debug purpose.
		*/
		real8 getUnitOffensiveCharge(sint4 x, sint4 y);

		/*
			Returns the enemy unit charge in a specific point
			for units in defensive mode.
			This method is only used in debug purpose.
		*/
		real8 getUnitDefensiveCharge(sint4 x, sint4 y);

		/*
			Returns the player base and units charge in a specific point.
			This method is only used in debug purpose.
		*/
		real8 getPlayerCharge(sint4 x, sint4 y);

		//Current tile width (in points) for charge map tiles
		static const int CMTILE_WIDTH;

		//The charge used for impassable terrain tiles
		static const int CHARGE_IMPASSABLE_TILES;

		//The size of the charge map
		int width;
		int height;

	private:

		//Charge map point
		struct point {
			real8 enemyBaseCharge_huntUnits;
			real8 enemyBaseCharge_huntBases;

			real8 enemyUnitOffensiveCharge;
			real8 enemyUnitDefensiveCharge;

			real8 playerBaseCharge;
			real8 playerUnitCharge;
		
			real8 neutralUnitCharge;

			real8 terrainCharge;
		};

		//The map
		point map[128][128];
};

#endif
