
#include "ChargeMapCreator.H"
#include "Map.H"
#include "PlayerInfo.H"
#include "StaticMap.H"
#include "Blackboard.H"

enum { TILE_UNKNOWN=0, TILE_WATER=1, TILE_GROUND=2, TILE_CLIFF=3 };

using namespace std;

bool is_initiated = false;
bool cm_thread_running = false;

/*
	Constructor.
*/
ChargeMapCreator::ChargeMapCreator() {
	
}

/*
	Destructor.
*/
ChargeMapCreator::~ChargeMapCreator() {
	
}

/*
	Returns true if the chargemap is ready for use, i.e. terrain and base charges are generated.
*/
bool ChargeMapCreator::isReady() {
	if (is_initiated && !cm_thread_running) {
		return true;
	}

	return false;
}

/*
	Returns false if the charge map needs to be initiated, i.e. the terrain and base
	charges must be generated.
*/
bool ChargeMapCreator::isInitiated() {
	if (!is_initiated && !cm_thread_running) {
		return false;
	}
	return true;
}

/*
	Returns a pointer to the charge map.
*/
ChargeMap* ChargeMapCreator::getMap() {
	return c_map;
}

/*
	Initiates the map. This method is threaded since it usually takes about 150 ms and each action
	frame is 125 ms.
*/
void ChargeMapCreator::initMap(GameStateModulePtr mState) {
	if (cm_thread_running) return;

	cm_thread_running = true;

	state = mState;

	playerId = state->get_game().get_client_player();
	enemyId = 0;
	if (playerId == 0) enemyId = 1;

	c_map = new ChargeMap(state);
	
	//Starts the SDL thread
	my_thread = SDL_CreateThread(ChargeMapCreator::execute, this);
}

/*
	Init thread run method.
*/
int ChargeMapCreator::execute(void *data) {
	ChargeMapCreator &cm = *static_cast<ChargeMapCreator*>(data);

	cout << "Starting map init thread" << endl;

	const Game& game(cm.state->get_game());

	clock_t start = clock();
	cout << "Creating terrain ChargeMap...";
	clock_t tStart = clock();
	cm.createTerrainMap(game);
	cm.postProcessTerrain2();
	cm.postProcessTerrain();
	cm.finalPostProcessTerrain();
	clock_t tEnd = clock();
	cout << "done in " << (tEnd - tStart) << " ms" << endl;

	cout << "Creating base objects ChargeMap...";
	clock_t bStart = clock();
	cm.createBaseMap(game);
	clock_t bEnd = clock();
	cout << "done in " << (bEnd - bStart) << " ms" << endl;

	clock_t end = clock();
	cout << "Map creation took " << (end - start) << " ms" << endl;

	is_initiated = true;

	cm_thread_running = false;

	cout << "Stopping map init thread" << endl;

	return 0;
}

/*
	Creates charge map for player and enemy bases.
*/
void ChargeMapCreator::createBaseMap(const Game &game) {
	int n = game.get_player_num();
	PlayerInfo& pi = game.get_cplayer_info();
	const sint4 cid = game.get_client_player();
	
	//Iterate through all players. The last player (n) is the server objects, but we dont
	//need to change charge for neutral server objects.
	for (int player = 0; player < n; player++) {
		const Game::ObjCont &objs = game.get_objs(player);

		bool friendly = false;
		if (player == cid) {
			friendly = true;
		}

		//Iterate through all objects for this player
		FORALL (objs, it) {
			GameObjPtr gob = (*it)->get_GameObj();
			if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
				//Valid object found
				uint4 objId = pi.get_id(gob);
				
				std::string objType = gob->bp_name();
				
				if (objType == "controlCenter") {
					if (friendly) {
						handlePlayerBase(gob);

						//Add the base to the blackboard
						int tileX = *gob->sod.x / ChargeMap::CMTILE_WIDTH;
						int tileY = *gob->sod.y / ChargeMap::CMTILE_WIDTH;
						Blackboard::addBase(tileX, tileY, objId);
					}
					else {
						handleEnemyBase_huntUnits(gob);
						handleEnemyBase_huntBases(gob);
					}
				}
			}
		}
	}
}

/*
	Updates the unit charge map to the current game state. Shall be called
	each action frame.
*/
void ChargeMapCreator::updateUnitMap(const Game& game) {
	c_map->clearUnitCharge();
	createUnitMap(game);
}

/*
	Updates the base charge map to the current game state. Shall be called
	when a enemy or friendly base is destroyed.
*/
void ChargeMapCreator::updateBaseMap(const Game& game) {
	c_map->clearBaseCharge();
	createBaseMap(game);
}

/*
	Creates charge map for player, enemy and neutral units.
*/
void ChargeMapCreator::createUnitMap(const Game &game) {
	int n = game.get_player_num();
	const sint4 cid = game.get_client_player();
	
	//Iterate through all players. The last player (n) is the server objects, but we dont
	//need to change charge for neutral server objects.
	for (int player = 0; player <= n; player++) {
		const Game::ObjCont &objs = game.get_objs(player);

		//Iterate through all objects for this player
		FORALL (objs, it) {
			GameObjPtr gob = (*it)->get_GameObj();
			if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
				//Valid object found
				
				bool friendly = false;
				if (player == cid) {
					friendly = true;
				}
				
				std::string objType = gob->bp_name();
				
				if (objType == "tank") {
					if (!friendly) {
						handleOffensiveTank(gob);
						handleDefensiveTank(gob);
					}
					else {
						handlePlayerTank(gob);
					}
				}
				if (objType == "marine") {
					if (!friendly) {
						handleOffensiveMarine(gob);
						handleDefensiveMarine(gob);
					}
					else {
						handlePlayerMarine(gob);
					}
				}
				if (objType == "sheep") {
					handleSheep(gob);
				}
			}
		}
	}
}

/*
	Update the charges for an enemy tank during offensive mode.
*/
void ChargeMapCreator::handleOffensiveTank(GameObjPtr gob) {
	int center_x = *gob->sod.x / ChargeMap::CMTILE_WIDTH;
	int center_y = *gob->sod.y / ChargeMap::CMTILE_WIDTH;

	int x1 = center_x - StaticMap::tankOffensive_w;
	int x2 = center_x + StaticMap::tankOffensive_w;
	int y1 = center_y - StaticMap::tankOffensive_w;
	int y2 = center_y + StaticMap::tankOffensive_w;
	
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			real8 charge = (real8)StaticMap::tankOffensive[cX - x1][cY - y1];
			c_map->updateEnemyUnitOffensiveCharge(cX, cY, charge);
		}
	}
}

/*
	Update the charges for an enemy marine during offensive mode.
*/
void ChargeMapCreator::handleOffensiveMarine(GameObjPtr gob) {
	int center_x = *gob->sod.x / ChargeMap::CMTILE_WIDTH;
	int center_y = *gob->sod.y / ChargeMap::CMTILE_WIDTH;

	int x1 = center_x - StaticMap::marineOffensive_w;
	int x2 = center_x + StaticMap::marineOffensive_w;
	int y1 = center_y - StaticMap::marineOffensive_w;
	int y2 = center_y + StaticMap::marineOffensive_w;
	
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			real8 charge = (real8)StaticMap::marineOffensive[cX - x1][cY - y1];
			c_map->updateEnemyUnitOffensiveCharge(cX, cY, charge);
		}
	}
}

/*
	Update the charges for an enemy tank during defensive mode.
*/
void ChargeMapCreator::handleDefensiveTank(GameObjPtr gob) {
	int center_x = *gob->sod.x / ChargeMap::CMTILE_WIDTH;
	int center_y = *gob->sod.y / ChargeMap::CMTILE_WIDTH;

	int x1 = center_x - StaticMap::tankDefensive_w;
	int x2 = center_x + StaticMap::tankDefensive_w;
	int y1 = center_y - StaticMap::tankDefensive_w;
	int y2 = center_y + StaticMap::tankDefensive_w;
	
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			real8 charge = (real8)StaticMap::tankDefensive[cX - x1][cY - y1];
			c_map->updateEnemyUnitDefensiveCharge(cX, cY, charge);
		}
	}
}

/*
	Update the charges for an enemy marine during defensive mode.
*/
void ChargeMapCreator::handleDefensiveMarine(GameObjPtr gob) {
	int center_x = *gob->sod.x / ChargeMap::CMTILE_WIDTH;
	int center_y = *gob->sod.y / ChargeMap::CMTILE_WIDTH;

	int x1 = center_x - StaticMap::marineDefensive_w;
	int x2 = center_x + StaticMap::marineDefensive_w;
	int y1 = center_y - StaticMap::marineDefensive_w;
	int y2 = center_y + StaticMap::marineDefensive_w;
	
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			real8 charge = (real8)StaticMap::marineDefensive[cX - x1][cY - y1];
			c_map->updateEnemyUnitDefensiveCharge(cX, cY, charge);
		}
	}
}

/*
	Appends a negative area where a player tank is moving.
*/
void ChargeMapCreator::appendDestinationForTank(int destX, int destY) {
	int x1 = destX - StaticMap::tankPlayer_w;
	int x2 = destX + StaticMap::tankPlayer_w;
	int y1 = destY - StaticMap::tankPlayer_w;
	int y2 = destY + StaticMap::tankPlayer_w;
	
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			real8 charge = (real8)StaticMap::tankPlayer[cX - x1][cY - y1];
			c_map->updatePlayerUnitCharge(cX, cY, charge);
		}
	}
}

/*
	Appends a negative area where a player marine is moving.
*/
void ChargeMapCreator::appendDestinationForMarine(int destX, int destY) {
	int x1 = destX - StaticMap::marinePlayer_w;
	int x2 = destX + StaticMap::marinePlayer_w;
	int y1 = destY - StaticMap::marinePlayer_w;
	int y2 = destY + StaticMap::marinePlayer_w;
	
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			real8 charge = (real8)StaticMap::marinePlayer[cX - x1][cY - y1];
			c_map->updatePlayerUnitCharge(cX, cY, charge);
		}
	}
}

/*
	Update the charges for a player tank.
*/
void ChargeMapCreator::handlePlayerTank(GameObjPtr gob) {
	int center_x = *gob->sod.x / ChargeMap::CMTILE_WIDTH;
	int center_y = *gob->sod.y / ChargeMap::CMTILE_WIDTH;

	int x1 = center_x - StaticMap::tankPlayer_w;
	int x2 = center_x + StaticMap::tankPlayer_w;
	int y1 = center_y - StaticMap::tankPlayer_w;
	int y2 = center_y + StaticMap::tankPlayer_w;
	
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			real8 charge = (real8)StaticMap::tankPlayer[cX - x1][cY - y1];
			c_map->updatePlayerUnitCharge(cX, cY, charge);
		}
	}
}

/*
	Update the charges for a player marine.
*/
void ChargeMapCreator::handlePlayerMarine(GameObjPtr gob) {
	int center_x = *gob->sod.x / ChargeMap::CMTILE_WIDTH;
	int center_y = *gob->sod.y / ChargeMap::CMTILE_WIDTH;

	int x1 = center_x - StaticMap::marinePlayer_w;
	int x2 = center_x + StaticMap::marinePlayer_w;
	int y1 = center_y - StaticMap::marinePlayer_w;
	int y2 = center_y + StaticMap::marinePlayer_w;
	
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			real8 charge = (real8)StaticMap::marinePlayer[cX - x1][cY - y1];
			c_map->updatePlayerUnitCharge(cX, cY, charge);
		}
	}
}

/*
	Update the charges for a neutral sheep.
*/
void ChargeMapCreator::handleSheep(GameObjPtr gob) {
	int center_x = *gob->sod.x / ChargeMap::CMTILE_WIDTH;
	int center_y = *gob->sod.y / ChargeMap::CMTILE_WIDTH;

	int x1 = center_x - StaticMap::sheep_w;
	int x2 = center_x + StaticMap::sheep_w;
	int y1 = center_y - StaticMap::sheep_w;
	int y2 = center_y + StaticMap::sheep_w;
	
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			real8 charge = (real8)StaticMap::sheep[cX - x1][cY - y1];
			c_map->updatePlayerUnitCharge(cX, cY, charge);
		}
	}
}

/*
	Update the charges a player base.
*/
void ChargeMapCreator::handlePlayerBase(GameObjPtr gob) {
	int center_x = *gob->sod.x / ChargeMap::CMTILE_WIDTH;
	int center_y = *gob->sod.y / ChargeMap::CMTILE_WIDTH;

	int x1 = center_x - StaticMap::basePlayer_w;
	int x2 = center_x + StaticMap::basePlayer_w;
	int y1 = center_y - StaticMap::basePlayer_w;
	int y2 = center_y + StaticMap::basePlayer_w;
	
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			real8 charge = (real8)StaticMap::basePlayer[cX - x1][cY - y1];
			c_map->updatePlayerBaseCharge(cX, cY, charge);
		}
	}

	//Clear terrain charges in an area close to the base
	int clearTerrainW = 2;
	int base_x1 = *gob->sod.x1 / ChargeMap::CMTILE_WIDTH - clearTerrainW;
	int base_x2 = *gob->sod.x2 / ChargeMap::CMTILE_WIDTH + clearTerrainW;
	int base_y1 = *gob->sod.y1 / ChargeMap::CMTILE_WIDTH - clearTerrainW;
	int base_y2 = *gob->sod.y2 / ChargeMap::CMTILE_WIDTH + clearTerrainW;

	for (int cX = base_x1; cX <= base_x2; cX++) {
		for (int cY = base_y1; cY <= base_y2; cY++) {
			c_map->setTerrainCharge(cX, cY, 0);
		}
	}
}

/*
	Update the charges for an enemy base.
*/
void ChargeMapCreator::handleEnemyBase_huntBases(GameObjPtr gob) {
	int center_x = *gob->sod.x / ChargeMap::CMTILE_WIDTH;
	int center_y = *gob->sod.y / ChargeMap::CMTILE_WIDTH;

	int x1 = center_x - StaticMap::base_w;
	int x2 = center_x + StaticMap::base_w;
	int y1 = center_y - StaticMap::base_w;
	int y2 = center_y + StaticMap::base_w;
	
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			//if (lineOfSightFree(cX, cY, gob)) {
				real8 charge = (real8)StaticMap::base[cX - x1][cY - y1];
				c_map->updateEnemyBaseCharge_huntBases(cX, cY, charge);
			//}
		}
	}
}

/*
	Update the charges for an enemy base.
*/
void ChargeMapCreator::handleEnemyBase_huntUnits(GameObjPtr gob) {
	int center_x = *gob->sod.x / ChargeMap::CMTILE_WIDTH;
	int center_y = *gob->sod.y / ChargeMap::CMTILE_WIDTH;

	int x1 = center_x - StaticMap::basePlayer_w;
	int x2 = center_x + StaticMap::basePlayer_w;
	int y1 = center_y - StaticMap::basePlayer_w;
	int y2 = center_y + StaticMap::basePlayer_w;
	
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			real8 charge = (real8)StaticMap::basePlayer[cX - x1][cY - y1];
			c_map->updateEnemyBaseCharge_huntUnits(cX, cY, charge);
		}
	}

	//Clear terrain charges in an area close to the base
	int clearTerrainW = 2;
	int base_x1 = *gob->sod.x1 / ChargeMap::CMTILE_WIDTH - clearTerrainW;
	int base_x2 = *gob->sod.x2 / ChargeMap::CMTILE_WIDTH + clearTerrainW;
	int base_y1 = *gob->sod.y1 / ChargeMap::CMTILE_WIDTH - clearTerrainW;
	int base_y2 = *gob->sod.y2 / ChargeMap::CMTILE_WIDTH + clearTerrainW;

	for (int cX = base_x1; cX <= base_x2; cX++) {
		for (int cY = base_y1; cY <= base_y2; cY++) {
			c_map->setTerrainCharge(cX, cY, 0);
		}
	}
}

/*
	Returns true if the line of sight from [tileX,tileY] to
	the GameObj is free from impassable terrain.

	Use this method with care since it is very slow.
*/
bool ChargeMapCreator::lineOfSightFree(int tileX, int tileY, GameObjPtr gob) {
	if (tileX < 0 || tileX >= c_map->width) return false;
	if (tileY < 0 || tileY >= c_map->height) return false;

	int useX = 0;
	int useY = 0;

	if (*gob->sod.shape == GameObj::RECTANGLE) {
		int tileX1 = *gob->sod.x1 / ChargeMap::CMTILE_WIDTH;
		int tileX2 = *gob->sod.x2 / ChargeMap::CMTILE_WIDTH;
		int tileY1 = *gob->sod.y1 / ChargeMap::CMTILE_WIDTH;
		int tileY2 = *gob->sod.y2 / ChargeMap::CMTILE_WIDTH;

		useX = tileX;
		if (useX < tileX1) useX = tileX1;
		if (useX > tileX2) useX = tileX2;

		useY = tileY;
		if (useY < tileY1) useY = tileY1;
		if (useY > tileY2) useY = tileY2;
	}
	if (*gob->sod.shape == GameObj::CIRCLE) {
		useX = *gob->sod.x / ChargeMap::CMTILE_WIDTH;
		useY = *gob->sod.y / ChargeMap::CMTILE_WIDTH;
	}

	int diffX = useX - tileX;
	int diffY = useY - tileY;
	int divider = max(abs(diffX), abs(diffY));

	int cnt = 0;
	int cX = tileX;
	int cY = tileY;

	while (cnt <= divider) {
		
		int mX = cX + calcNext(cnt, diffX, divider);
		int mY = cY + calcNext(cnt, diffY, divider);

		real8 charge = c_map->getTerrainCharge(mX, mY);
		if (charge <= -35) {
			//Impassable tile found
			return false;
		}

		cnt++;
	}

	return true;
}

/*
	Help method for lineOfSightFree().
*/
int ChargeMapCreator::calcNext(int cnt, int diffVal, int divider) {
	real8 next = (real8) cnt * (real8)diffVal / (real8)divider;

	int val = (int)(next * 10);
	int res = (int)next;

	if (val % 10 >= 5) res++;

	return res;
}

/*
	Creates charge map for the terrain.
*/
void ChargeMapCreator::createTerrainMap(const Game &game) {
	const Map<GameTile> &gameMap = game.get_map();

	//Iterate through the map tiles
	FORS (y, gameMap.get_height()) {
		FORS (x, gameMap.get_width()) {
			const GameTile &tile = gameMap(x,y);

			//Start and end points for this tile in our charge map
			int cmTilesPerGameTile = game.get_tile_points() / ChargeMap::CMTILE_WIDTH;

			int startX = x * cmTilesPerGameTile;
			int endX = startX + cmTilesPerGameTile;
			int startY = y * cmTilesPerGameTile;
			int endY = startY + cmTilesPerGameTile;

			if (tile.get_split() == Tile::NO_SPLIT) {
				//No split. Tile only has one terrain type

				if (tile.get_typeW() != TILE_GROUND) {
					for (int cX = startX; cX < endX; cX++) {
						for (int cY = startY; cY < endY; cY++) {
							handleImpassableTerrainTile(cX, cY);
						}
					}
				}
			}
			else if (tile.get_split() == Tile::BT_SPLIT) {
				//Tile is split |\|
				if (tile.get_typeW() != TILE_GROUND) {
					for (int cX = startX; cX < endX; cX++) {
						for (int cY = startY; cY < endY; cY++) {
							if (y >= x) {
								handleImpassableTerrainTile(cX, cY);
							}
						}
					}
				}
				else if (tile.get_typeE() != TILE_GROUND) {
					for (int cX = startX; cX < endX; cX++) {
						for (int cY = startY; cY < endY; cY++) {
							if (y <= x) {
								handleImpassableTerrainTile(cX, cY);
							}
						}
					}
				}
				else {
					cout << "Unknown BT split" << endl;
				}
			}
			else if (tile.get_split() == Tile::TB_SPLIT) {
				//Tile is split |/|

				if (tile.get_typeW() != TILE_GROUND) {
					for (int cX = startX; cX < endX; cX++) {
						for (int cY = startY; cY < endY; cY++) {
							if (y <= (15 - x)) {
								handleImpassableTerrainTile(cX, cY);
							}
						}
					}
				}
				else if (tile.get_typeE() != TILE_GROUND) {
					for (int cX = startX; cX < endX; cX++) {
						for (int cY = startY; cY < endY; cY++) {
							if (y >= (15 - x)) {
								handleImpassableTerrainTile(cX, cY);
							}
						}
					}
				}
				else {
					cout << "Unknown TB split" << endl;
				}
			}
			else {
				cout << "Unknown tile split" << endl;
			}
		}
	}
}

/*
	This method can be called after creating terrain map to
	clear paths between blocked terrain.
*/
void ChargeMapCreator::postProcessTerrain() {
	for (int cX = 0; cX < c_map->width; cX++) {
		for (int cY = 0; cY < c_map->width; cY++) {
			real8 t_charge = c_map->getTerrainCharge(cX, cY);
			if (t_charge >= TBConfig::params.ChargeMap_PathMinThreshold && t_charge <= TBConfig::params.ChargeMap_PathMaxThreshold) {
				if (clearTile(cX, cY, t_charge)) {
					c_map->setTerrainCharge(cX, cY, 0);
					//c_map->setTerrainCharge(cX, cY, 100); //Uncomment for easier debug
				}
			}
		}
	}
}

/*
	This method can be called after creating terrain map to
	block bad areas in the terrain (i.e. areas surrounded by
	impassable terrain).
*/
void ChargeMapCreator::postProcessTerrain2() {
	real8 preFill = 200;

	for (int cX = 0; cX < c_map->width; cX++) {
		for (int cY = 0; cY < c_map->width; cY++) {
			real8 t_charge = c_map->getTerrainCharge(cX, cY);

			if (t_charge > ChargeMap::CHARGE_IMPASSABLE_TILES && t_charge < -5) {
				int blockCount = 0;
				
				//Passable tile. Check surroundings.

				//Row 1
				if (isBlocked(cX, cY, -1, -1, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;
				if (isBlocked(cX, cY, -0.5, -1, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;
				if (isBlocked(cX, cY, 0, -1, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;
				if (isBlocked(cX, cY, 0.5, -1, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;
				if (isBlocked(cX, cY, 1, -1, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;

				//Row 2
				if (isBlocked(cX, cY, -1, -0.5, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;
				if (isBlocked(cX, cY, -1, 0, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;
				if (isBlocked(cX, cY, -1, 0.5, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;

				if (isBlocked(cX, cY, 1, -0.5, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;
				if (isBlocked(cX, cY, 1, 0, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;
				if (isBlocked(cX, cY, 1, 0.5, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;

				//Row 3
				if (isBlocked(cX, cY, -1, 1, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;
				if (isBlocked(cX, cY, -0.5, 1, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;
				if (isBlocked(cX, cY, 0, 1, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;
				if (isBlocked(cX, cY, 0.5, 1, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;
				if (isBlocked(cX, cY, 1, 1, TBConfig::params.ChargeMap_Fill_DistToCheck)) blockCount++;

				//Check if enough directions is blocked => preFill
				if (blockCount >= TBConfig::params.ChargeMap_Fill_MaxBlockedDirections) {
					c_map->setTerrainCharge(cX, cY, preFill);
				}
			}
		}
	}

	//Change preFill values to impassable tile charge
	for (int cX = 0; cX < c_map->width; cX++) {
		for (int cY = 0; cY < c_map->width; cY++) {
			real8 t_charge = c_map->getTerrainCharge(cX, cY);
			if (t_charge == preFill) {
				c_map->setTerrainCharge(cX, cY, ChargeMap::CHARGE_IMPASSABLE_TILES);
			}
		}
	}
}

/*
	Help method for postProcessTerrain2().
	Returns true of the direction [dX,dY] from [cX,cY] has an impassable tile
	within a specified distance (iterations).
*/
bool ChargeMapCreator::isBlocked(int cX, int cY, real8 dX, real8 dY, int iterations) {
	int cnt = 0;
	real8 mX = (real8)cX;
	real8 mY = (real8)cY;

	while (cnt < iterations) {
		real8 t_charge = c_map->getTerrainCharge((int)mX, (int)mY);
		if (t_charge <= ChargeMap::CHARGE_IMPASSABLE_TILES) {
			//Blocked
			return true;
		}

		mX += dX;
		mY += dY;
		cnt++;
	}

	return false;
}

/*
	This is the final postprocess method for terrain map creation.
	It adds an impassable tile charge to the outer bounds of the map.
*/
void ChargeMapCreator::finalPostProcessTerrain() {
	for (int cX = 0; cX < c_map->width; cX++) {
		c_map->setTerrainCharge(cX, 0, ChargeMap::CHARGE_IMPASSABLE_TILES);
		c_map->setTerrainCharge(cX, c_map->height - 1, ChargeMap::CHARGE_IMPASSABLE_TILES);
	}
	for (int cY = 0; cY < c_map->height; cY++) {
		c_map->setTerrainCharge(0, cY, ChargeMap::CHARGE_IMPASSABLE_TILES);
		c_map->setTerrainCharge(c_map->width - 1, cY, ChargeMap::CHARGE_IMPASSABLE_TILES);
	}
}

/*
	Returns true if the terrain tile at the specific point shall be cleared.
	A tile is cleared if it belongs to a path between blocked terrain.
	The charge parameter is the original terrain charge in the specific point.
*/
bool ChargeMapCreator::clearTile(int cX, int cY, real8 charge) {
	int x1 = cX;
	int y1 = cY;
	int x2 = cX;
	int y2 = cY;
	
	x1 = cX - 1;
	y1 = cY;
	x2 = cX + 1;
	y2 = cY;
	if (checkSurroundings(x1, y1, x2, y2, charge)) {
		return true;
	}

	x1 = cX;
	y1 = cY - 1;
	x2 = cX;
	y2 = cY + 1;
	if (checkSurroundings(x1, y1, x2, y2, charge)) {
		return true;
	}

	return false;
}

/*
	Help method for clearTile().
	Returns true if a tile shall be cleared based on how the charges in
	[x1,y1] and [x2,y2] differs from the charge parameter.
*/
bool ChargeMapCreator::checkSurroundings(int x1, int y1, int x2, int y2, real8 charge) {
	real8 ch1 = c_map->getTerrainCharge(x1, y1);
	real8 ch2 = c_map->getTerrainCharge(x2, y2);

	if (ch1 >= TBConfig::params.ChargeMap_PathMinThreshold && ch1 <= TBConfig::params.ChargeMap_PathMaxThreshold) {
		if (ch2 >= TBConfig::params.ChargeMap_PathMinThreshold && ch2 <= TBConfig::params.ChargeMap_PathMaxThreshold) {
			if (ch1 < charge && ch2 < charge) {
				return true;
			}
		}
	}

	return false;
}

/*
	Update the charges for an impassable terrain tile at [cX,cY]
*/
void ChargeMapCreator::handleImpassableTerrainTile(int cX, int cY) {
	int x1 = cX - StaticMap::terrain_w;
	int x2 = cX + StaticMap::terrain_w;
	int y1 = cY - StaticMap::terrain_w;
	int y2 = cY + StaticMap::terrain_w;
	
	for (int mX = x1; mX <= x2; mX++) {
		for (int mY = y1; mY <= y2; mY++) {
			
			if (mX == cX && mY == cY) {
				c_map->updateTerrainCharge(cX, cY, ChargeMap::CHARGE_IMPASSABLE_TILES);
			}
			else {
				real8 charge = (real8)StaticMap::terrain[mX - x1][mY - y1];
				c_map->updateTerrainCharge(mX, mY, charge);
			}
		}
	}
}
