#include "ClientGUI.H"
#include "SDL_gfxPrimitives.h"

using namespace std;

bool is_gui_initiated = false;

const int ClientGUI::MODE_OFFENSIVE = 1;
const int ClientGUI::MODE_DEFENSIVE = 2;
const int ClientGUI::MODE_BASE = 3;
const int ClientGUI::MODE_OFFENSIVE_UNITS = 4;
const int ClientGUI::MODE_DEFENSIVE_UNITS = 5;
const int ClientGUI::MODE_TERRAIN = 6;
const int ClientGUI::MODE_PLAYER = 7;

/*
	Constructor.
*/
ClientGUI::ClientGUI() {
	screen = 0;
	gfx_mag_frac = 16;
	quit = false;
	writeStatus = false;

	mode = ClientGUI::MODE_OFFENSIVE;
}

/*
	Returns true if the GUI is ready (init finished).
*/
bool ClientGUI::isReady() {
	return is_gui_initiated;
}

/*
	Initiates the GUI.
*/
void ClientGUI::init() {
	s_x = 780;
	s_y = 780;
	adj_x = s_x/32;
	adj_y = s_y/32;

	// open screen
    
	// Have a preference for 8-bit, but accept any depth

	screen = SDL_SetVideoMode(s_x, s_y, 0, SDL_HWSURFACE | SDL_ANYFORMAT);
	if (screen == 0) {
		//   cerr << "Couldn't set " << s_x << "x" << s_y << "x8 video mode: " << SDL_GetError() << endl;
		exit(1);
	}

	ostringstream oss;
	oss << "Tankbattle";

	SDL_WM_SetCaption(oss.str().c_str(), 0);

	is_gui_initiated = true;

	gui_running = false;
}

/*
	Updates the ChargeMap.
*/
void ClientGUI::updateRef(ChargeMap* map) {
	c_map = map;
}

/*
	Starts the auto update thread.
*/
void ClientGUI::startThread() {
	gui_running = true;
	my_thread = SDL_CreateThread(ClientGUI::execute, this);
}

int ClientGUI::execute(void *data) {
	ClientGUI &cg = *static_cast<ClientGUI*>(data);

	while (!cg.quit) {
		cg.display();

		SDL_Delay(250);
	}

	return -1;
}

/*
	Updates the GUI.
*/
void ClientGUI::display() {
	if (SDL_MUSTLOCK(screen) && SDL_LockSurface(screen) < 0 ) {
		fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
		return;
	}

	//Set title
	if (mode == ClientGUI::MODE_OFFENSIVE) {
		ostringstream oss;
		oss << "Enemy bases and units - Offensive mode";
		SDL_WM_SetCaption(oss.str().c_str(), 0);
	}
	if (mode == ClientGUI::MODE_DEFENSIVE) {
		ostringstream oss;
		oss << "Enemy bases and units - Defensive mode";
		SDL_WM_SetCaption(oss.str().c_str(), 0);
	}
	if (mode == ClientGUI::MODE_BASE) {
		ostringstream oss;
		oss << "Enemy bases";
		SDL_WM_SetCaption(oss.str().c_str(), 0);
	}
	if (mode == ClientGUI::MODE_OFFENSIVE_UNITS) {
		ostringstream oss;
		oss << "Enemy units - Offensive mode";
		SDL_WM_SetCaption(oss.str().c_str(), 0);
	}
	if (mode == ClientGUI::MODE_DEFENSIVE_UNITS) {
		ostringstream oss;
		oss << "Enemy units - Defensive mode";
		SDL_WM_SetCaption(oss.str().c_str(), 0);
	}
	if (mode == ClientGUI::MODE_TERRAIN) {
		ostringstream oss;
		oss << "Terrain";
		SDL_WM_SetCaption(oss.str().c_str(), 0);
	}
	if (mode == ClientGUI::MODE_PLAYER) {
		ostringstream oss;
		oss << "Player bases and units";
		SDL_WM_SetCaption(oss.str().c_str(), 0);
	}
	
	//Update GFX
	for (int cX = 0; cX < c_map->width; cX++) {
		for (int cY = 0; cY < c_map->height; cY++) {
			real8 charge = 0;

			//Get correct charge for the current mode
			if (mode == ClientGUI::MODE_OFFENSIVE) {
				charge = c_map->getOffensiveCharge(cX, cY);
			}
			if (mode == ClientGUI::MODE_DEFENSIVE) {
				charge = c_map->getDefensiveCharge(cX, cY);
			}
			if (mode == ClientGUI::MODE_BASE) {
				charge = c_map->getEnemyBaseCharge(cX, cY);
			}
			if (mode == ClientGUI::MODE_OFFENSIVE_UNITS) {
				charge = c_map->getUnitOffensiveCharge(cX, cY);
			}
			if (mode == ClientGUI::MODE_DEFENSIVE_UNITS) {
				charge = c_map->getUnitDefensiveCharge(cX, cY);
			}
			if (mode == ClientGUI::MODE_TERRAIN) {
				charge = c_map->getTerrainCharge(cX, cY);
			}
			if (mode == ClientGUI::MODE_PLAYER) {
				charge = c_map->getPlayerCharge(cX, cY);
			}
			
			int w = 6;
			SDL_Rect rect;
			rect.x = cX * w;
			rect.y = cY * w;
			rect.w = w;
			rect.h = w;

			SDL_FillRect(screen, &rect, getColor(charge));
		}
	}

	if (SDL_MUSTLOCK(screen)) SDL_UnlockSurface(screen);

	//SDL_Flip(screen);
	SDL_UpdateRect(screen, 0, 0, 0, 0);
}

/*
	Returns a display color for the specified charge.
*/
Uint32 ClientGUI::getColor(real8 charge) {
	//Update
	if (charge <= ChargeMap::CHARGE_IMPASSABLE_TILES / 2) {
		//Impassable terrain
		RGB rgb(44, 44, 44);
		return rgb.color();
	}
	if (charge == 0) {
		RGB rgb(0, 0, 0);	
		return rgb.color();
	}
	
	if (charge < 0) {
		int R = (int)((float)(charge * -1) * 6);
		int G = (int)(R / 4);
		int B = (int)(R / 4);

		if (R < 1) R = 1;
		if (R > 254) R = 254;
		if (G < 1) G = 1;
		if (G > 254) G = 254;
		if (B < 1) B = 1;
		if (B > 254) B = 254;

		RGB rgb(R, G, B);
		return rgb.color();
	}
	if (charge > 0) {
		int B = (int)((float)charge * 2);
		int R = (int)(B / 4);
		int G = (int)(B / 4);

		if (R < 1) R = 1;
		if (R > 254) R = 254;
		if (G < 1) G = 1;
		if (G > 254) G = 254;
		if (B < 1) B = 1;
		if (B > 254) B = 254;
		
		RGB rgb(R, G, B);
		return rgb.color();
	}
	
	return 0;
}

static bool effective_key(sint4 sym, sint4 mod) {
  mod = 0;
  sint4 ignore[] = { SDLK_LSHIFT, SDLK_RSHIFT, SDLK_LCTRL, SDLK_RCTRL,
		     SDLK_LALT, SDLK_RALT,  SDLK_LMETA, SDLK_RMETA };

  FORU (i, sizeof(ignore)/sizeof(ignore[0])) {
    if (sym == ignore[i]) return false;
  }

  return true;
}

/*
	Event handler for key events in the GUI window.
*/
void ClientGUI::event() {
	SDL_Event event;
	SDL_EnableUNICODE(true);

	static sint4 dx = 0, dy = 0;

	while (SDL_PollEvent(&event)) {

		if (event.type == SDL_KEYDOWN || event.type == SDL_KEYUP) {

			//      cout << "%" << flush;
			bool up = event.type == SDL_KEYUP;
			char ch;
      
			if ((event.key.keysym.unicode & 0xFF80) == 0) {
				ch = event.key.keysym.unicode & 0x7F;
				//cout << "ch= " << (int)ch << endl;
			}
			else {
				printf("An International Character.\n");
			}
      
			bool update = false;

			sint4 sym = event.key.keysym.sym;
			sint4 mod = event.key.keysym.mod;
			if (effective_key(sym, mod)) {
				switch (sym) {
					case 49: mode = ClientGUI::MODE_OFFENSIVE; update = true; break;
					case 50: mode = ClientGUI::MODE_DEFENSIVE; update = true; break;
					case 51: mode = ClientGUI::MODE_BASE; update = true; break;
					case 52: mode = ClientGUI::MODE_OFFENSIVE_UNITS; update = true; break;
					case 53: mode = ClientGUI::MODE_DEFENSIVE_UNITS; update = true; break;
					case 54: mode = ClientGUI::MODE_TERRAIN; update = true; break;
					case 55: mode = ClientGUI::MODE_PLAYER; update = true; break;
					case 56: writeStatus = true; break;
					case 27: quit = true; break; // ESC -> quit
					case SDLK_LEFT:  dx = -1; break;
					case SDLK_RIGHT: dx = +1; break;
					case SDLK_UP:    dy = -1; break;
					case SDLK_DOWN:  dy = +1; break;
				}

				if (up) dx = dy = 0;
			}
			if (event.type == SDL_QUIT) { 
				if (gui_running) {
					SDL_KillThread(my_thread);
				}
				quit = true;
				break; 
			}

			if (update) {
				cout << "Change view detected" << endl;
				display();
				SDL_Delay(50);
			}

			if (quit) {
				cout << "Quit event detected" << endl;
			}
		}
	}  
}
