#ifndef FCStructure_H
#define FCStructure_H

#include "GameObj.H"
#include "GameStateModule.H"
#include "FCStructureItem.H"

/*
	This is the main class for the data structure used in fire coordination.

	The basic structure is an array with friendly units that can fire at the y axis
	and possible enemy targets at the x axis, and 1 for all possible attack combinations
	and 0 for combinations where attacking isn't possible.
	It is implemented as a vector of vectors to get a more dynamic and memoryefficient structure.
*/
class FCStructure {
	
	typedef GameObj* GameObjPtr;
	typedef FCStructureItem* FCStructureItemPtr;
	
	public:

		/*
			Clears the structure. Should be called before a new coordination begins.
		*/
		void clear();

		/*
			Returns true if at least one possible attack possibility is added to the structure.
		*/
		bool hasEntries();
		
		/*
			Adds a possible attack combination. The unit [unit] can attack [target].
		*/
		void addAttackPossibility(uint4 unitId, GameObjPtr unit, uint4 targetId, GameObjPtr target);

		/*
			After all attack possibilities has been added to the structure this method shall
			be called to do the actual fire coordination.
			After processing each friendly unit has 0 or 1 targets to fire at.
		*/
		void findAttackScheme();

		/*
			Returns the object Id of the target for the specified unit. Returns -1 if no targets are available.
		*/
		int getTargetId(uint4 unitId);

		/*
			Returns the GameObj of the target for the specified unit. Returns null if no targets are available.
		*/
		GameObj* getTargetObj(uint4 unitId);
		
		/*
			Dumps the data structure to the console window for debugging purpose.
		*/
		void dump();

	private:

		/*
			If there are more attacking units than what is needed to destroy the
			target this method removes the unnecessary attacking units.
		
			To decide which entries to remove the method calculates the number of occurences
			the attacker unit has in the whole data structure, and the ones with max
			occurences (which are most likely to be in other combinations) are removed.
		*/
		void removeUnnecessary(FCStructureItemPtr item, int noToRemove);

		/*
			Calculates the number of occurences an attacking unit has in the whole
			data structure.
		*/
		int calculateOccurences(uint4 unitId);
		
		/*
			Removes targets that during or after coordination processing aren't attacked by any units.
		*/
		void removeEmptyEntries();

		/*
			This method is used during coordination processing.
			The logic has decided that the specified unit/target combination (FCStructureItem) shall remain after
			the coordination is finished and this method removes the attacking units from other unit/target combinations.
		*/
		void removeAttackers(FCStructureItemPtr item);

		/*
			Removes the unit [unitId] from attacking other targets than [currentTargetId].
		*/
		void removeAttacker(uint4 currentTargetId, uint4 unitId);

		//A list of unit/target combinations (i.e. which units can attack which target)
		vector<FCStructureItem> items;
};

#endif
