
#include "TBEventHandler.H"
#include "GameStateModule.H"
#include "GfxModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameTile.H"
#include "ServerObjData.H"
#include "Options.H"
#include "UnitAgent.H"
#include "Blackboard.H"

using namespace std;

bool TBEventHandler::handle_event(const Event& e) {
	if (e.get_who() == GameStateModule::FROM) {
    
		if (e.get_what() == GameStateModule::STOP_MSG) {
			cout << "Game stopped" << endl;
			exit(0);
		}
    
		if (e.get_what() == GameStateModule::READ_ERROR_MSG) {
			cout << "Error occured: " << state.error << endl;
			exit(1);
		}
    
		if (e.get_what() == GameStateModule::VIEW_MSG) {
      
			compute_actions();
			// Send my actions
			state.gsm->send_actions();
			
			Game &game = state.gsm->get_game();	

			// gfx and frame statistics
      
			if (state.gfxm) state.gfxm->process_changes();
			
			sint4 vf = game.get_view_frame();
			sint4 af = game.get_action_frame();
			sint4 sa = game.get_skipped_actions();
			sint4 ma = game.get_merged_actions();

			//cout << "[frame " << vf << "]" << endl;
	
			if (vf != af || sa || ma > 1) {
				/*cout << "frame " << vf;
				if (af < 0) cout << " [no action]";
				else if (vf != af) cout << " [behind by " << (vf - af) << " frame(s)]";
				if (sa) cout << " [skipped " << sa << "]";
				if (ma > 1) cout << " [merged " << ma << "]";
				cout << endl;*/
			
				/*if (af != 0 && vf != af) cout << " [behind by " << (vf - af) << " frame(s)]" << endl;
				if (sa) cout << " [skipped " << sa << "]" << endl;
				if (ma > 1) cout << " [merged " << ma << "]" << endl;*/
			}

			if (state.gfxm) {
	
				// don't draw if we are far behind
				if (abs(vf-af) < 10) {
					state.gfxm->draw();
					state.just_drew = true;
				}
			}
			    
			return true;
		}
	}
		    
	return false;
}


void TBEventHandler::compute_actions() {
	const GameChanges &changes = state.gsm->get_changes();

	if (!changes.new_tile_indexes.empty()) {
		// New tiles only occurs if we have fog of war
		//cout << "#new tiles = " << changes.new_tile_indexes.size() << endl;
	}
  
	if (!changes.new_boundaries.empty()) {
		// New boundaries only occurs if we have fog of war
		//cout << "#new boundaries = " << changes.new_boundaries.size() << endl;
	}

	if (!changes.new_objs.empty()) {
		// New objects only happens if we have fog of war or if units can be built
		//cout << "#new objects = " << changes.new_objs.size() << endl;
	}

	if (!changes.changed_objs.empty()) {
		// Changed objects. When do this happen?
		//cout << "#changed objects = " << changes.changed_objs.size() << endl;
	}

	if (!changes.vanished_objs.empty()) {
		// Vanished objects. Can only happen if we have fog of war
		//cout << "#vanished objects = " << changes.vanished_objs.size() << endl;
	}

	//Start time for this action frame
	clock_t startTime = clock();

	bool baseDestroyed = false;
	bool tankDestroyed = false;

	if (!changes.dead_objs.empty()) {
		FORALL (state.gsm->get_changes().dead_objs, it) {
        	GameObj *gob = (*it)->get_GameObj();
			std::string objType = gob->bp_name();
			//cout << "Dead object: " << objType << endl;
			if (objType == "controlCenter") {
				//A base has been destroyed
				baseDestroyed = true;
			}
			if (objType == "tank") {
				//A tank has been destroyed
				tankDestroyed = true;
			}
		}
	}

	//Current game state
	const Game &game = state.gsm->get_game();
	
	//This is the unique id for this player
	const sint4 cid = game.get_client_player();

	const Game::ObjCont &objs = game.get_objs(cid);
	
	bool debug;
	Options::get("-debug", debug);

	if (!mapCreator.isInitiated()) {
		//We need to init the mapCreator
		mapCreator.initMap(state.gsm);
	}

	if (debug && mapCreator.isReady() && !gui.isReady()) {
		//We need to init the GUI
		gui.init();
		gui.updateRef(mapCreator.getMap());
		gui.startThread();
	}

	if (mapCreator.isReady()) {
		//Update game tactics
		Blackboard::updateState(game);

		//If a base has been destroyed, we need to update the base map
		if (baseDestroyed) {
			mapCreator.updateBaseMap(game);
		}
		if (tankDestroyed) {
			Blackboard::updateAliveStatus(game, cid);
		}
		
		//Everything is ready. Update the unit map each frame

		//clock_t s = clock();
		mapCreator.updateUnitMap(game);
		//clock_t e = clock();
		//cout << "Update unit map took " << (e - s) << " ms" << endl;

		if (debug && gui.isReady()) {
			gui.updateRef(mapCreator.getMap());
		}
	}

	if (debug && gui.isReady()) {
		gui.event();
		if (gui.quit) exit(0);
	}
	
	//Iterate through this players' objects
	if (mapCreator.isReady()) {
		//Coordinate fire at enemy targets
		coordinator.coordinate_fire(state.gsm, &mapCreator);

		FORALL (objs, it) {
			GameObj *gob = (*it)->get_GameObj();
			if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
				std::string objType = gob->bp_name();

				//This GameObj is in the game
				if (objType == "tank") {
					UnitAgent::compute_actions(gob, state.gsm, &mapCreator, &coordinator);
				}
				if (objType == "marine") {
					UnitAgent::compute_actions(gob, state.gsm, &mapCreator, &coordinator);
				}
			}
		}
	}

	clock_t endTime = clock();
	sint4 af = game.get_action_frame();

	//Write unit/frame status information
	/*if (mapCreator.isReady() && (gui.writeStatus || af % 100 == 0)) {
		double elapsed = ((double)((endTime - startTime) * 1000))/CLOCKS_PER_SEC;

		cout << "" << endl;
		cout << "AF=" << af << " execTime=" << elapsed << " playerId=" << mapCreator.playerId << endl;
		
		if (gui.writeStatus) {
			Blackboard::printDebug();
		}

		gui.writeStatus = false;
	}*/

	//This writes unit status necessary for the GA optimization. It can be commented out in the final version
	/*if (mapCreator.isReady() && (baseDestroyed || af % 10 == 0)) {
		Blackboard::printUnitStatus(game);
	}*/
	
	//cout << "Iteration time: " << elapsed << endl;
}
