#include "UnitAgent.H"
#include "StaticMap.H"
#include "Blackboard.H"
#include "TBConfig.H"
#include "FireCoordinator.H"

const int UnitAgent::TANK_FIRERANGE = 112;
const int UnitAgent::TANK_HP = 150;
const int UnitAgent::TANK_ARMOR = 1;
const sint4 UnitAgent::TANK_MIN_DMG = 26;
const sint4 UnitAgent::TANK_MAX_DMG = 34;

const int UnitAgent::MARINE_FIRERANGE = 64;
const int UnitAgent::MARINE_HP = 80;
const int UnitAgent::MARINE_ARMOR = 1;
const sint4 UnitAgent::MARINE_MIN_DMG = 5;
const sint4 UnitAgent::MARINE_MAX_DMG = 7;

const int UnitAgent::CONTROLCENTER_HP = 1200;
const int UnitAgent::CONTROLCENTER_ARMOR = 0;

/*
	Compute actions for this tank object.
*/
void UnitAgent::compute_actions(GameObjPtr unit, GameStateModulePtr state, ChargeMapCreatorPtr mapCreator, FireCoordinatorPtr coordinator) {
	Game& game = state->get_game();
	
	uint4 objId = game.get_cplayer_info().get_id(unit);
	int unitX = *unit->sod.x / ChargeMap::CMTILE_WIDTH;
	int unitY = *unit->sod.y / ChargeMap::CMTILE_WIDTH;

	//Get which mode (offensive/defensive) to use
	bool offensive = getMode(unit);

	//Fire if any targets are within range
	/*//Old version with local optimization
	bool fired = canFire(unit, unitX, unitY, objId, state, offensive, mapCreator);
	if (fired) {
		return;
	}*/
	//New version with global coordination
	bool fired = canFireNew(unit, unitX, unitY, objId, state, offensive, coordinator);
	if (fired) {
		return;
	}

	//Check if we need to do a random "restless" move
	bool rndmoved = doRestlessMove(unit, unitX, unitY, objId, offensive, mapCreator);
	if (rndmoved) return;

	//No targets to fire at, check if we can move
	bool moved = canMove(unit, unitX, unitY, objId, offensive, mapCreator);
	if (moved) return;
	
	//Cannot fire and no better position found, stay at the same position
	Blackboard::setAction(unitX, unitY, objId, Blackboard::getType(unit), Blackboard::AC_NO_ACTION, Blackboard::NO_ID, Blackboard::NO_ID, Blackboard::NO_ID, offensive);
}

/*
	Checks if the unit needs to do a random "restless" move. Returns false if no restless move is needed or possible.
*/
bool UnitAgent::doRestlessMove(GameObjPtr unit, int unitX, int unitY, uint4 objId, bool offensive, ChargeMapCreatorPtr mapCreator) {
	//First, check if unit is restless. If not return false.
	if (!Blackboard::isRestless(objId)) {
		return false;
	}

	ChargeMapPtr chargeMap = mapCreator->getMap();

	int direction = Blackboard::getDirection(objId);
	if (direction == 0) {
		//No direction. Randomize
		Random rand;
		rand.srand((uint4)clock_t());
		direction = rand.ranged_uint4(1, 8);
	}

	int destX = unitX;
	int destY = unitY;
	int dist = 2;

	if (direction == 1) {
		destX = unitX - dist;
		destY = unitY - dist;
	}
	else if (direction == 2) {
		destX = unitX;
		destY = unitY - dist;
	}
	else if (direction == 3) {
		destX = unitX + dist;
		destY = unitY - dist;
	}
	else if (direction == 4) {
		destX = unitX + dist;
		destY = unitY;
	}
	else if (direction == 5) {
		destX = unitX + dist;
		destY = unitY + dist;
	}
	else if (direction == 6) {
		destX = unitX;
		destY = unitY + dist;
	}
	else if (direction == 7) {
		destX = unitX - dist;
		destY = unitY + dist;
	}
	else if (direction == 8) {
		destX = unitX - dist;
		destY = unitY;
	}
	else {
		cout << "[" << objId << "] Invalid random direction: " << direction << endl;
	}

	if (canDoMove(unitX, unitY, destX, destY, chargeMap)) {
		//We've found a free position to move to!
		//Set direction
		Blackboard::setDirection(objId, direction);
	}
	else {
		//Unable to move
		//Clear direction
		Blackboard::setDirection(objId, 0);
	}

	if (destX != unitX || destY != unitY) {
		//Move to new position
		int actualX = destX * ChargeMap::CMTILE_WIDTH + ChargeMap::CMTILE_WIDTH / 2;
		int actualY = destY * ChargeMap::CMTILE_WIDTH + ChargeMap::CMTILE_WIDTH / 2;

		int speed = *unit->sod.max_speed;
		Vector<sint4> params;
		params.push_back(actualX);
		params.push_back(actualY);
		params.push_back(speed);

		unit->set_action("move", params);

		Blackboard::setAction(unitX, unitY, objId, Blackboard::getType(unit), Blackboard::AC_MOVE, destX, destY, Blackboard::NO_ID, offensive);
		
		//Add some negative charge to this tanks destination so other
		//units are less likely to move to the same spot
		if (Blackboard::getType(unit) == Blackboard::TYPE_TANK) {
			mapCreator->appendDestinationForTank(destX, destY);
		}
		if (Blackboard::getType(unit) == Blackboard::TYPE_MARINE) {
			mapCreator->appendDestinationForMarine(destX, destY);
		}

		//Unit has moved. Return true
		return true;
	}
	
	//Unable to move. Return false
	return false;
}

/*
	Checks if the unit can move to the specified position.
*/
bool UnitAgent::canDoMove(int fromX, int fromY, int toX, int toY, ChargeMapPtr chargeMap) {
	//First, check boundaries
	if (toX < 0 || toY < 0) return false;
	if (toX >= chargeMap->width || toY >= chargeMap->height) return false;
	
	int diffX = getDiff(fromX, toX);
	int diffY = getDiff(fromY, toY);

	//No need to check the start position since the unit is already there -> it is passable
	int cX = fromX + diffX;
	int cY = fromY + diffY;
	bool end = false;

	while (!end) {
		real8 charge = chargeMap->getMinimalTerrainCharge(cX, cY);
		if (charge <= ChargeMap::CHARGE_IMPASSABLE_TILES / 2) {
			//Impassable tile
			return false;
		}

		if (cX == toX && cY == toY) {
			//We're done!
			end = true;
		}

		cX += diffX;
		cY += diffY;
	}

	return true;
}

/*
	Gets the diff when moving from fromPos to toPos. Diff is -1, 0 or 1.
*/
int UnitAgent::getDiff(int fromPos, int toPos) {
	int diff = toPos - fromPos;
	if (diff < 0) diff = -1;
	if (diff > 0) diff = 1;
	return diff;
}

/*
	Checks if there are better positions nearby to move to.
	Returns true if unit moves, false if not.
*/
bool UnitAgent::canMove(GameObjPtr unit, int unitX, int unitY, uint4 objId, bool offensive, ChargeMapCreatorPtr mapCreator) {
	//If no targets are within range, check if there are better positions to move to
	ChargeMapPtr chargeMap = mapCreator->getMap();

	//Calculate which tiles to check
	int x1 = max(unitX - TBConfig::params.UnitAgent_LookAheadTiles, 0);
	int x2 = min(unitX + TBConfig::params.UnitAgent_LookAheadTiles, chargeMap->width - 1);
	int y1 = max(unitY - TBConfig::params.UnitAgent_LookAheadTiles, 0);
	int y2 = min(unitY + TBConfig::params.UnitAgent_LookAheadTiles, chargeMap->height - 1);

	//Iterate through the tiles to get the highest (=best) charge
	//Start with the charge for current tile
	real8 bestCharge = chargeMap->getCharge(unitX, unitY, offensive) - getUnitCharge(unitX, unitY, unitX, unitY);
	int bestX = unitX;
	int bestY = unitY;
	
	//Get the surrounding tile with highest charge
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			if (cX != unitX || cY != unitY) {
				if (canDoMove(unitX, unitY, cX, cY, chargeMap)) {
					real8 cCharge = chargeMap->getCharge(cX, cY, offensive) - getUnitCharge(cX, cY, unitX, unitY);
					if (cCharge > bestCharge) {
						bestCharge = cCharge;
						bestX = cX;
						bestY = cY;
					}
				}
			}
		}
	}

	//If we've found a surrounding tile with higher charge than current tile, move there
	if (bestX != unitX || bestY != unitY) {
		int actualX = bestX * ChargeMap::CMTILE_WIDTH + ChargeMap::CMTILE_WIDTH / 2;
		int actualY = bestY * ChargeMap::CMTILE_WIDTH + ChargeMap::CMTILE_WIDTH / 2;

		int speed = *unit->sod.max_speed;
		Vector<sint4> params;
		params.push_back(actualX);
		params.push_back(actualY);
		params.push_back(speed);

		unit->set_action("move", params);

		Blackboard::setAction(unitX, unitY, objId, Blackboard::getType(unit), Blackboard::AC_MOVE, bestX, bestY, Blackboard::NO_ID, offensive);
		
		//Add some negative charge to this tanks destination so other
		//units are less likely to move to the same spot
		mapCreator->appendDestinationForTank(bestX, bestY);
		
		//Unit has moved
		return true;
	}

	//Unit has not moved
	return false;
}

/*
	Decides which mode (offensive = true, defensive = false) to use for a tank.
*/
bool UnitAgent::getMode(GameObjPtr unit) {
	bool offensive = true;

	//Check 1: If the tank has low HP it enters defensive mode
	int unitHP = unit->get_int("hp");
	real8 maxHP = (real8)UnitAgent::TANK_HP;
	if (Blackboard::getType(unit) == Blackboard::TYPE_MARINE) {
		maxHP = (real8)UnitAgent::MARINE_HP;
	}

	real8 left = (real8)unitHP / maxHP * (real8)100;
	if (left < TBConfig::params.UnitAgent_HPDefensiveLimit) {
		offensive = false;
	}
	
	return offensive;
}

/*
	Gets the charge this tank generates at a point. It is used to remove
	the negative impact the tank generates in the map when calculating the
	actions for this frame.
*/
real8 UnitAgent::getUnitCharge(int x, int y, int centerX, int centerY) {
	int diffX = x - centerX;
	int diffY = y - centerY;

	real8 charge = (real8)StaticMap::tankPlayer[StaticMap::tankPlayer_w - diffX][StaticMap::tankPlayer_w - diffY];

	return charge;
}

/*
	Checks for enemy targets within range and fires if any targets found.
	Returns true if fire at a target.

	This method uses a global target coordination to decide what target to attack (if more than one is within firerange).
*/
bool UnitAgent::canFireNew(GameObjPtr unit, int unitX, int unitY, uint4 objId, GameStateModulePtr state, bool offensive, FireCoordinatorPtr coordinator) {
	Game& game = state->get_game();
	
	//First, check time to shoot
	sint4 lastFired = unit->component("weapon")->get_int("last_fired");
	if (lastFired >= 0) {
		uint4 game_tick = game.tick();
		sint4 cooldown = unit->component("weapon")->get_int("cooldown");
		int timeToShoot = cooldown - (game_tick - (uint4)lastFired);

		if (timeToShoot > 0) {
			//Cooldown period. Not allowed to shoot
			if (!Blackboard::isLonely(unitX, unitY, objId)) {
				//Unit is allowed to move during cooldown if it isn't lonely
				return false;
			}
			else {
				//If we return true the unit stays at this position until the target is destroyed
				//-> less offensive behaviour
				Blackboard::setAction(unitX, unitY, objId, Blackboard::getType(unit), Blackboard::AC_AWAIT_COOLDOWN, Blackboard::NO_ID, Blackboard::NO_ID, Blackboard::NO_ID, offensive);
				return true;
			}
		}
	}

	//Get the target for this unit from the global coordinator
	//-1 = No target available.
	int targetId = coordinator->getTargetId(objId);
	GameObjPtr target = coordinator->getTargetObj(objId);
	if (targetId >= 0) {
		Vector<sint4> args;
		args.push_back(targetId);
		unit->component("weapon")->set_action("attack", args);

		int targetX = *target->sod.x / ChargeMap::CMTILE_WIDTH;
		int targetY = *target->sod.y / ChargeMap::CMTILE_WIDTH;

		Blackboard::setAction(unitX, unitY, objId, Blackboard::getType(unit), Blackboard::AC_ATTACK_MARINE, targetX, targetY, game.get_cplayer_info().get_id(target), offensive);

		return true;
	}

	return false;
}

/*
	Checks for enemy targets within range and fires if any targets found.
	Returns true if fire at a target.

	If this method is used, each unit decides for themselves what target to attack (if more than one is within firerange).
*/
bool UnitAgent::canFire(GameObjPtr unit, int unitX, int unitY, uint4 objId, GameStateModulePtr state, bool offensive, ChargeMapCreatorPtr mapCreator) {
	Game& game = state->get_game();
	PlayerInfo& pi = game.get_cplayer_info();
	
	//First, check time to shoot
	sint4 lastFired = unit->component("weapon")->get_int("last_fired");
	if (lastFired >= 0) {
		uint4 game_tick = game.tick();
		sint4 cooldown = unit->component("weapon")->get_int("cooldown");
		int timeToShoot = cooldown - (game_tick - (uint4)lastFired);

		if (timeToShoot > 0) {
			//Cooldown period. Not allowed to shoot
			if (!Blackboard::isLonely(unitX, unitY, objId)) {
				//Unit is allowed to move during cooldown if it isn't lonely
				return false;
			}
			else {
				//If we return true the unit stays at this position until the target is destroyed
				//-> less offensive behaviour
				Blackboard::setAction(unitX, unitY, objId, Blackboard::getType(unit), Blackboard::AC_AWAIT_COOLDOWN, Blackboard::NO_ID, Blackboard::NO_ID, Blackboard::NO_ID, offensive);
				return true;
			}
		}
	}
	
	const Game::ObjCont &objs = game.get_objs(mapCreator->enemyId);	

	real8 bestAttackValue = -100;
	uint4 bestObjectId = 0;
	GameObj* bestObj;

	//Iterate through enemy objects
	FORALL (objs, it) {
		GameObj *gob = (*it)->get_GameObj();
		if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
			real8 dist = unit->distance_to(*gob);

			real8 fireRange = TANK_FIRERANGE;
			if (Blackboard::getType(unit) == Blackboard::TYPE_MARINE) {
				fireRange = MARINE_FIRERANGE;
			}

			if (dist <= fireRange) {
				std::string objType = gob->bp_name();
				
				//Check visibility
				bool visible = game.can_see_obj(*unit, *gob);
				
				if (visible) {
					//Object is within range and is visible, calculate attack value
					real8 cAttackValue = getAttackValue(unitX, unitY, offensive, gob, game.get_cplayer_info().get_id(gob), dist, (int)fireRange, mapCreator);

					if (cAttackValue > bestAttackValue) {
						bestAttackValue = cAttackValue;
						bestObjectId = pi.get_id(gob);
						bestObj = gob;
					}
				}
			}
		}
	}

	if (bestAttackValue >= 0.1) {
		//We've found a target!

		Vector<sint4> args;
		args.push_back(bestObjectId);
		unit->component("weapon")->set_action("attack", args);

		int targetX = *bestObj->sod.x / ChargeMap::CMTILE_WIDTH;
		int targetY = *bestObj->sod.y / ChargeMap::CMTILE_WIDTH;

		std::string objType = bestObj->bp_name();
		if (objType == "tank") {
			Blackboard::setAction(unitX, unitY, objId, Blackboard::getType(unit), Blackboard::AC_ATTACK_TANK, targetX, targetY, game.get_cplayer_info().get_id(bestObj), offensive);
		}
		else if (objType == "marine") {
			Blackboard::setAction(unitX, unitY, objId, Blackboard::getType(unit), Blackboard::AC_ATTACK_MARINE, targetX, targetY, game.get_cplayer_info().get_id(bestObj), offensive);
		}
		else if (objType == "controlCenter") {
			Blackboard::setAction(unitX, unitY, objId, Blackboard::getType(unit), Blackboard::AC_ATTACK_BASE, targetX, targetY, game.get_cplayer_info().get_id(bestObj), offensive);
		}
		else {
			cout << "Unknown attack target:" << objType << endl;
		}

		return true;
	}

	return false;
}

/*
	Calculates the attack value for a target. If the value is high, the target
	is important to fire at (if more than one target are within range for
	a player unit).
*/
real8 UnitAgent::getAttackValue(int unitX, int unitY, bool offensive, GameObjPtr enemy, uint4 enemyId, real8 distance, int max_fireRange, ChargeMapCreatorPtr mapCreator) {
	std::string objType = enemy->bp_name();

	//Step 1: Calculate a value based on distance to target
	real8 distMultiplier = (real8)5;
	real8 distVal = distMultiplier * ((real8)max_fireRange - distance) / (real8)max_fireRange;
	
	//Step 2: Calculate a valued based on how many units
	//attacking the target and current target HP.
	real8 attacksMultiplier = (real8)10;
	real8 unitMultiplier = (real8)20;
	real8 attacksVal = 0;
	if (objType == "tank") {
		attacksVal = attacksMultiplier * Blackboard::getTankAttackMultiplier(enemy, enemyId) * unitMultiplier;
	}
	if (objType == "marine") {
		attacksVal = attacksMultiplier * Blackboard::getMarineAttackMultiplier(enemy, enemyId) * unitMultiplier;
	}
	if (objType == "controlCenter") {
		if (Blackboard::state == Blackboard::STATE_HUNT_UNITS && mapCreator->getMap()->canSpotEnemyUnit(unitX, unitY, offensive)) {
			//Dont attack bases until most of enemy units are destroyed
			return -100;
		}

		attacksVal = attacksMultiplier * Blackboard::getBaseAttackMultiplier(enemy, enemyId);
	}

	real8 attackValue = attacksVal;
	if (attacksVal > 0) {
		attackValue += distVal;
	}

	return attackValue;
}
