#ifndef UnitAgent_H
#define UnitAgent_H

#include "ChargeMap.H"
#include "ChargeMapCreator.H"
#include "GameObj.H"
#include "Game.H"
#include "GameStateModule.H"
#include "FireCoordinator.H"

/*
	This class represents an autonomous agent for units. It compute the actions
	for a tank or marine based on the current game state, charge map and firecoordinator.
*/
class UnitAgent {
	
	typedef ChargeMap* ChargeMapPtr;
	typedef GameObj* GameObjPtr;
	typedef ChargeMapCreator* ChargeMapCreatorPtr;
	typedef GameStateModule* GameStateModulePtr;
	typedef FireCoordinator* FireCoordinatorPtr;

	public:

		/*
			Compute actions for this tank object.
		*/
		static void compute_actions(GameObjPtr unit, GameStateModulePtr state, ChargeMapCreatorPtr mapCreator, FireCoordinatorPtr coordinator);

		/*
			Decides which mode (offensive = true, defensive = false) to use for a tank.
		*/
		static bool getMode(GameObjPtr unit);

		/*
			Checks if the unit needs to do a random "restless" move. Returns false if no restless move is needed or possible.
		*/
		static bool doRestlessMove(GameObjPtr unit, int unitX, int unitY, uint4 objId, bool offensive, ChargeMapCreatorPtr mapCreator);

		/*
			Checks if the unit can do a random move to the specified position.
		*/
		static bool canDoMove(int fromX, int fromY, int toX, int toY, ChargeMapPtr chargeMap);

		/*
			Gets the diff when moving from fromPos to toPos. Diff is -1, 0 or 1.
		*/
		static int getDiff(int fromPos, int toPos);

		/*
			Checks for enemy targets within range and fires if any targets found.
			Returns true if fire at a target.

			If this method is used, each unit decides for themselves what target to attack (if more than one is within firerange).
		*/
		static bool canFire(GameObjPtr unit, int unitX, int unitY, uint4 objId, GameStateModulePtr state, bool offensive, ChargeMapCreatorPtr mapCreator);

		/*
			Checks for enemy targets within range and fires if any targets found.
			Returns true if fire at a target.

			This method uses a global target coordination to decide what target to attack (if more than one is within firerange).
		*/
		static bool canFireNew(GameObjPtr unit, int unitX, int unitY, uint4 objId, GameStateModulePtr state, bool offensive, FireCoordinatorPtr coordinator);

		/*
			Checks if there are better positions nearby to move to.
			Returns true if unit moves, false if not.
		*/
		static bool canMove(GameObjPtr unit, int unitX, int unitY, uint4 objId, bool offensive, ChargeMapCreatorPtr mapCreator);

		/*
			Calculates the attack value for a target. If the value is high, the target
			is important to fire at (if more than one target are within range for
			a player unit).
		*/
		static real8 getAttackValue(int unitX, int unitY, bool offensive, GameObjPtr enemy, uint4 enemyId, real8 distance, int max_fireRange, ChargeMapCreatorPtr mapCreator);

		/*
			Gets the charge this tank generates at a point. It is used to remove
			the negative impact the tank generates in the map when calculating the
			actions for this frame.
		*/
		static real8 getUnitCharge(int x, int y, int centerX, int centerY);

		//Max firerange for tanks
		static const int TANK_FIRERANGE;
		//Max firerange for marines
		static const int MARINE_FIRERANGE;
		//Max HP for tanks
		static const int TANK_HP;
		//Armor for tanks
		static const int TANK_ARMOR;
		//Max HP for marines
		static const int MARINE_HP;
		//Armor for marins
		static const int MARINE_ARMOR;
		//Max HP for control centers
		static const int CONTROLCENTER_HP;
		//Armor for control centers
		static const int CONTROLCENTER_ARMOR;

		//Min/max damage for tank attacks
		static const sint4 TANK_MIN_DMG;
		static const sint4 TANK_MAX_DMG;
		//Min/max damage for marine attacks
		static const sint4 MARINE_MIN_DMG;
		static const sint4 MARINE_MAX_DMG;
};

#endif
