/* Updated Dr. Buro's robin.c for scheduling the round robin tournament 
 * 
 * Krysta Mirzayans  11-24/05/06
 * this version completes the first incomplete round before going on to
 * a new round.
 */

#include <string>
#include <sstream>
#include <map>        
#include <vector>
#include <cstdio>
#include <fstream>
#include <cstdlib>
#include <iostream>

//#include "TGlobal.H"   //this in turn needs a lot of other files, so perhaps add to the main makefile instead

using namespace std;

#define true      1
#define false     0
#define FOR(i,n)  for (i=0; i < (n); i++)
#define FOREVER   for (;;)
#define NO_RES 0        //For tournament matrix display: 
                        //0-only display result, 1- display everything other than result

struct GAME
{       
  int opponent;         //index of opponent
  bool valid;           //are the results of this game valid - has the game been played?
  string result;        //arbitrary result string
  int seed;             //game (map) seed
  bool seen;            //flag for display purposes
  int   dummy;          //true <=> PASS game (indicates a BYE)
};

struct PLAYERDATA
{ 
  int index;            //player index 
  string name;          //player name
  int seed;             //player's seed value 
  int points;           //player's score ???? may not need this       
  vector<string> comps; //computers that this player has
};


/**
 * Error reporting function, which displays the given error
 * message to standard out and terminates the program with code 20.
 *
 * @param msg  error message to be displayed
 */
void Error(string msg)
{
  if(msg == "usage")
    {
      cerr<< "usage: updateRobin -help | [-s player1 player2] configFile\n\
 -help   : print a detailed description                                            \n\
 -single : single mode fills any holes in previous rounds between these two players \n\n";
      exit(20);
    }
  cerr << "*** " << msg << endl;
  
  exit(20);
}


/**
 * Helper function for parsing the configuration file which
 * skips sections of comments (denoted by #).
 * 
 * @param in  the input file stream (configuration file)
 * @return    a bool flag indicating whether or not end-of-file is reached (true - eof, false - otherwise)
 */
bool skipComments(ifstream &in)
{
  string s;
  char c;
  FOREVER
    {
      if((c = in.peek())== EOF) return true;
      if((c != '#') && (c != ' ') && (c != '\n')) break;   	
      getline(in, s);
    }
  return false;
}


/**
 * Generates the tournament matrix by scheduling playernum-1 rounds
 * matrix structure:
 *          Player0  Player1
 * rnd0 gm0
 *      gm1
 * rnd1 gm0
 *      gm1
 *
 * @param matrix          the data structure that holds the schedule of games
 * @param playernum       the total number of players (including a ghost player if the actual number of players is odd)
 * @param pass            a flag indicating whether or not a ghost player exists (true - there is a ghost player)
 * @param games_per_pair  the number of games to be played per pair of players (in a given round)
 * @param game_seed       the base seed for game maps found by ex-or'ing all the players' seeds. Each game's seed is 
 *                        found by incrementing this base seed.
 * @param single          flag indicating whether one or two games are played per map (true-one game, false-two games)
 */
void generate(GAME **matrix, int playernum, bool pass, int games_per_pair, int game_seed, bool single)
{
  int i, j, k, startround = 0, round;
  FOR (i, playernum-1)
    {
      round = startround;

      FOR (j, playernum-1)
        {
          //If you're on the diagonal, either pair with the last player or give
          //a bye if there's an odd number of players (playernum-1 is pass, if odd)
	  FOR(k, games_per_pair)
	    {
	      
	      if (i == j)
		{
		  
		  //play against the last player:
		  matrix[round*games_per_pair + k][j].opponent = playernum - 1;
		  matrix[round*games_per_pair + k][playernum-1].opponent = i;

		  //no results are yet valid:
		  matrix[round*games_per_pair + k][j].valid = false;
		  matrix[round*games_per_pair + k][playernum-1].valid = false;

		  //set the default results   ***COME BACK: does it matter to what?????
		  matrix[round*games_per_pair + k][j].result = "NULL";
		  matrix[round*games_per_pair + k][playernum-1].result = "NULL";
		  
		  //set the game seed:
		  matrix[round*games_per_pair + k][j].seed = game_seed;
		  matrix[round*games_per_pair + k][playernum-1].seed = game_seed;

		  //set the seen flag:
		  matrix[round*games_per_pair + k][j].seen = false;
		  matrix[round*games_per_pair + k][playernum-1].seen = false;

		  //dummy game if the last player is pass:
		  matrix[round*games_per_pair + k][j].dummy = pass;
		  matrix[round*games_per_pair + k][playernum-1].dummy = pass;
		}
	      else
		{ //If not on the diagonal, can't be dummy (since limits don't include playernum-1)
		  matrix[round*games_per_pair + k][j].opponent = i;
		  matrix[round*games_per_pair + k][j].valid = false;
		  matrix[round*games_per_pair + k][j].result = "NULL";
		  matrix[round*games_per_pair + k][j].dummy = false;
		  matrix[round*games_per_pair + k][j].seen = false;
		  //set the game seed
		  matrix[round*games_per_pair + k][j].seed = game_seed;
		}
	      
	      if(single || (k&1)) 
		++game_seed;
	    }

	  round++;
          if (round >= playernum-1) round = 0; //because of wrapping to mix up
                                               //who plays in which round
        }
      
      startround ++;                           //to mix up playing orders
    }
}


/**
 * Displays the help message to standard out when -help is the command line argument.
 * Exits the program with code 10.
 */
void print_help()
{
  cout << "updateRobin - a round-robin ORTS tournament manager.         \n";
  cout << "It reads a list of players and a list of games played so far \n\
and schedules/runs the first incomplete round of the tournament. \n\
A round-by-round tournament table and results are printed.              \n";
  cout << "\nTournament-file syntax:\n";
  cout << "# comment                                                    \n\
\n# player section                                                      \n\
\n<number-of-games-per-pair>                                            \n\
\ns|d # (single: one game per map or double: two games per map)         \n\
\n<name> <rng-seed> <computer1> <computer2> ...                         \n\
\n...                                                                   \n\
\n%                                                                     \n\
\n# game section                                                        \n\
\nr<round> g<game> <map-seed> <name1> <name2> <arbitrary-result-string> \n\
\n...                                                                   \n";
  cout << "\nSample tournament-file:\n";
  cout << "\n#header comments                                           \n\
\n4   # 4 games per pair                                                \n\
d   # double: two games / map                                           \n\
\nUofA  23411832 ug01 ug02                                              \n\
USC   41253533 ug03 ug04                                                \n\
UMich 23435223 ug05 ug06                                                \n\
\n%                                                                     \n\
\n# games                                                               \n\
\n# game seed for map i computed as XOR(all player seeds) + i           \n\
# if double, maps are used for two games                                \n\
\nr1 g1 234512345 UofA  USC   0 1 0 3  2 0 2.5 0                        \n\
r1 g2 234512345 USC   UofA  0 1 0 3  2 0 2.5 0                          \n\
r1 g3 234512346 UofA  USC   0 1 0 3  2 0 2.5 0                          \n\
r1 g4 234512346 USC   UofA  0 1 0 3  2 0 2.5 0                          \n\
\nr2 g1 234512345 UofA  UMich 0 1 0 3  2 0 2.5 0                        \n\
r2 g2 234512345 UMich UofA  0 1 0 3  2 0 2.5 0                          \n\
r2 g3 234512346 UofA  UMich 0 1 0 3  2 0 2.5 0                          \n\
r2 g4 234512346 UMich UofA  0 1 0 3  2 0 2.5 0                          \n\
\nr3 g1 234512345 USC   UMich 0 1 0 3  2 0 2.5 0                        \n\
r3 g2 234512345 UMich USC   0 1 0 3  2 0 2.5 0                          \n\
r3 g3 234512346 USC   UMich 0 1 0 3  2 0 2.5 0                          \n\
r3 g4 234512346 UMich USC   0 1 0 3  2 0 2.5 0                          \n\n"; 

  exit(10);
}


/**
 * Displays the scheduled game matrix in the following form to standard out:
 *         Player0  Player1
 * rnd0 gm0
 *      gm1
 * rnd1 gm0
 *      gm1
 *
 * If preprocessor flag NO_RES is set, displays all info other than results for each game.
 * Otherwise, displays only the results for each game.(this flag will be removed later on)
 *
 * @param matrix          the data structure that holds the schedule of games
 * @param playernum       the total number of players (including a ghost player if the actual number of players is odd)
 * @param games_per_pair  the number of games to be played per pair of players (in a given round)
 * @param single          flag indicating whether one or two games are played per map (true-one game, false-two games)
 */
void displayMatrix(GAME **matrix, int playernum, int games_per_pair, bool single)
{
  cout << "GENERATED TOURNAMENT MATRIX:\n";
  cout << "(seed, dummy?, valid?, opponent?-first player implicitely column index, result)\n";
  int i, j;
  cout << endl;
  FOR(i, playernum)
    cout << "\tplayer " << i << ":";
  cout << endl;

  FOR(i, (playernum-1)*games_per_pair)
    {
      if(!(i % games_per_pair))
	cout << "\nRound " << i/games_per_pair << ":";
      else
	cout << "\t";
      FOR(j, playernum)
	{
#if NO_RES
	  cout << matrix[i][j].seed << " " << matrix[i][j].dummy << " " << matrix[i][j].valid << " " << matrix[i][j].opponent << "\t";
#else 
	  cout << matrix[i][j].result << "\t";
#endif
	}
      cout << endl;
    }
  cout << endl;
} 

/**
 * Displays all players' data to standard out,  as read in from the configuration file.
 *
 * @param single          flag indicating whether one or two games are played per map (true-one game, false-two games)
 * @param games_per_pair  the number of games to be played per pair of players (in a given round)
 * @param playernum       the total number of players (including a ghost player if the actual number of players is odd)
 * @param numCompsEach    the number of computers each player has to play on
 * @param players         map of PLAYERDATA indexed by player index
 */
void displayPlayers(bool single, int games_per_pair, int playernum, int numCompsEach, map<int, PLAYERDATA*> players) 
{
  int i, j;
  cout << "\nAll info obtained about players from config file:\n\n";
  cout << "single? " << single << "\tgames_per_pair: " << games_per_pair << endl << endl;
  FOR(i, playernum)
    {
      cout << "Name:\t" << players[i]->name << endl;
      cout << "Seed:\t" << players[i]->seed << endl;
      cout << "Index:\t" << players[i]->index << endl;
      cout << "Computers:\t";
      FOR(j, numCompsEach)
	cout << players[i]->comps[j] << " ";
      cout << endl << endl;
    }
}


/**
 * Converts everything after the first character of the given string
 * parameter to an integer (ie "r1" to 1). If these characters do not
 * represent integers, calls (@link Error) function which exits the program. 
 *
 * @param name  string whose contents following the first character
 *              is to be converted to an integer value (ie "r1" to 1)  
 * @return      integer value represented by the contents of the string
 *              parameter following the first character.
 */
int stringToInt(string name)
{
  int temp = 0;
  int num = 0;
  for(unsigned int i=1; i < name.length(); ++i)   //put uint here!!!!!!!!
    {
      num *= 10;
      if(!(temp = name[i] - '0')) Error("incorrect format of game string");
      num += temp;
    }
 return num;
}

/**
 * Plays the given game of the given round for a given player. Makes the ssh call to
 * the server and stores the result of this call as this game's result. (not fully implemented yet)
 *
 * @param matrix          the data structure that holds the schedule of games
 * @param index           the column index of the matrix, which corresponds to one player of this game
 * @param roundNum        the index of the given round (indexing starts at 0)
 * @param games_per_pair  the number of games to be played per pair of players (in a given round)
 * @param gameNum         the index of the given game in this round (indexing starts at 0)
 */
void play(GAME **matrix, int index, int roundNum, int games_per_pair, int gameNum)
{
  string s = "";
  char c;
  FILE *fpREAD =popen("ssh krysta@ug01.cs.ualberta.ca ./WORK/PLAYER ","r");
  
  while(1) 
    {
      c = fgetc(fpREAD);
      if(c==EOF || c=='\n') break;  //***COME BACK: in reality, it may be wrong to ignore \n!!!(~not simply a string)  
      s += c;
    }
  fclose(fpREAD);
  
  cout << " result: _" << s << "_" << endl;
  matrix[roundNum*games_per_pair+gameNum][index].result = s;
  matrix[roundNum*games_per_pair+gameNum][matrix[roundNum*games_per_pair+gameNum][index].opponent].result = s;
}

/**
 * Main function
 */
int main(int argc, char **argv)
{
  int argi, i, j;
  string s;
  char c;

  int playernum = 0;                //total number of players(including a ghost player)
  bool pass = false;                //if true, indicates there is a ghost player
  int games_per_pair;               //how many games must be played between each pair of players
  bool single = true;               //1(if true) or 2(if false) games per map
  int game_seed;                    //the map seed for each game
  int numCompsEach = 0;             //number of computers per player
  bool opt_s = false;               //single mode: fill in holes for given two players 
  
  if (argc < 2) 
    Error("usage");

  argi = 1;                         //start with the first command line argument
  if (!strcmp(argv[argi], "-help")) print_help();
  map<int, PLAYERDATA*> players;    //indexed starting from 0 
  
  //***COME BACK: ignored the command line argument bit for now!!!!!
  //***COME BACK: assumed the input is perfect: -help or the configuration file name!!!!!
  int pair[] = { -1, -1 };                //for single mode, these are the indices of the players we're dealing with
  string names[] = { "", "" };
  FOREVER                     //***COME BACK: add more command line options
    {
      if(argi >= argc) Error("usage"); 
      if(!strcmp(argv[argi], "-single"))
	{ 
	  opt_s = true;
	  FOR(i, 2)
	    {
	      ++argi;
	      if(argi >= argc) Error("usage");
	      names[i] = argv[argi];
	    }
	  ++argi;   
	  continue; 
	} //the following two args should be player names
      break;
    }

  //PARSING THE INPUT CONFIGURATION FILE:
  
  ifstream in(argv[argi]);
  if(!in.is_open()) Error("unable to open the tournament file for reading");
  if(skipComments(in)) Error("end of file?");
 
  //games_per_pair:
  if(!(in >> games_per_pair)) Error("improper file format -- games per pair?");
  if(skipComments(in)) Error("end of file");

  //single or double games per map:
  getline(in,s);
  switch(s[0])     //junk the rest of the line
    {              //***COME BACK: could do a comment check after the first character!!!!!
    case 's':
      break;
    case 'd':
      single = false;
      break;
    default:
      Error("improper file format -- single or double?");
    }
  
  //players and player information (format: name seed computer1 computer2 ...):
  if(skipComments(in)) Error("end of file?");
  string computers;
  istringstream iss;
  FOREVER                          //read the first player's information
    {
      if(!getline(in, s)) Error("end of file?");
      if(!s.empty())
	{
	  if(s[0]=='%') break;     //% is end of player section marker 
	  if(s[0] != '#')          //# is comment marker    
	    {
	      iss.str(s);
	      players[playernum] = new PLAYERDATA;
	      (players[playernum])->index = playernum;
 	
	      if(!(iss >> (players[playernum])->name)) Error("player name?");

	      //if single mode, get the index of the pair of players:
	      if(opt_s)
		FOR(i,2)
		  if(names[i]==players[playernum]->name) pair[i]=playernum;
	      
	      if(!(iss >> (players[playernum])->seed)) Error("player seed?");
	      game_seed = players[playernum]->seed; //initialize the game seed
	      FOREVER
		{
		  if(!(iss >> computers))
		    break;
		  if(computers[0] == '#')
		    break;
		  ((players[playernum])->comps).push_back(computers);
		  ++numCompsEach;
		}
	      if(numCompsEach == 0) Error("Players need at least one computer each");   	  
	      playernum++;
	      break;
	    }  
	}
    }
 


  if(skipComments(in)) Error("end of file?");
  FOREVER                          //read the rest of the player's information 
    {
      if((c = in.peek()) == EOF) Error("end of file?");
      in >> s;
      if(!s.empty())
	{ 
	  if(s[0] == '%') break;
	  if(s[0] != '#')
	    {
	      players[playernum] = new PLAYERDATA;
	      players[playernum]->index = playernum;
	      players[playernum]->name = s; 
	      //if single mode, get the index of the pair of players:
	      if(opt_s)
		FOR(i,2)
		  if(names[i]==players[playernum]->name) pair[i]=playernum;

	      if(!(in >> players[playernum]->seed)) Error("player seed?");
	      game_seed ^= players[playernum]->seed;  //XOR all the players' seeds to get the game seed
    
	      FOR(i, numCompsEach)
		{
		  if(!(in >> computers)) Error("player computers?");
		  players[playernum]->comps.push_back(computers);
		}
	      playernum++;
	    }
	}
    }

  //if single mode, check if our players are valid: 
  if(opt_s)
    FOR(i,2)
      if(pair[i]<0)
	{
	  cout << i << " " << pair[i] << endl; 
	  Error("no player " + names[i] + " in this tournament");
	}
  
  //Print out all the read-in player information  
  displayPlayers(single, games_per_pair, playernum, numCompsEach, players); 
    
  //if playernum is odd, add a ghost player
  if(playernum & 1)
    {
      players[playernum] = new PLAYERDATA;
      players[playernum]->index = playernum;
      players[playernum]->name = "PASS"; 
      players[playernum]->seed = 0;
      pass = true;                           //there is a ghost player now
      
      //***COME BACK: how do you initialize the comps (does it even matter)?????
      FOR(i, numCompsEach)
	{
	  players[playernum]->comps.push_back("NULL");
	}
      ++playernum;
    }

  //Generate the schedule in the tournament matrix:
  GAME **matrix = new GAME*[(playernum-1)*games_per_pair];
  FOR(i, (playernum-1)*games_per_pair)
    matrix[i] = new GAME[playernum];
  generate(matrix, playernum, pass, games_per_pair, game_seed, single);
  //Print out the schedule
  displayMatrix(matrix, playernum, games_per_pair, single);
  
  //NOW CONSIDER GAME RESULTS: 
  //if none listed in config file, it is first round, else fill in the tournament state thus far
  
  //***COME BACK:
  //PROBLEM: the results at the end of the sample format file are not scheduled in the way that 
  //         this algorithm schedules, so for a certain round/game, it expects certain players to
  //         play each other. do you have to handle if unexpected pairings of players occurs in the 
  //         config file since the results appended will be generated by this algorithm anyway??? 
  
  int roundNum = 0;
  int gameNum = 0;
  bool roundflag = false;
  bool gameflag = false;
  string player1, player2;  //***COME BACK: this assumes a two player game -- but game type1 is single player!!!!!
  int index1 = -1, index2 = -1;
  int temp_r;
  if(!skipComments(in))     //if not EOF (ie. if there is a result section) 
    {
      FOREVER
	{
	  if(!(in >> s)) break;  //suffices for eof check?
	  if(!s.empty())
	    {
	      if(s[0] != '#')
		{
		  if(!roundflag)
		    {
		      if(s[0] != 'r') Error("game format: round number");
		      temp_r = stringToInt(s) - 1;
		      if(temp_r > playernum-1) Error("game format: too many rounds");
		      if(temp_r > roundNum)  roundNum = temp_r; 

		      roundflag = true;
		    }
		  else if(roundflag && !gameflag)
		    {
		      if(s[0] != 'g') Error("game format: game number");
		      gameNum = stringToInt(s) - 1;
		      if(gameNum >= games_per_pair) Error("game format: too many games per pair");
		     
		      gameflag = true;
		    }
		  if(roundflag && gameflag)//if already have round & game number
		    {
		 
		      if(!(in >> game_seed)) Error("game format: game seed");
		      if(!(in >> player1)) Error("game format: player1 name");
		      if(!(in >> player2)) Error("game format: player2 name");
  
		      FOR(i, playernum)
			{
			  if(players[i]->name == player1)
			    index1 = players[i]->index;
			  if(players[i]->name == player2)
			    index2 = players[i]->index;
			}
		      if(index1 == index2) Error("game format: same player names");
		      if(index1 < 0 || index2 < 0) Error("game format: invalid player name");
		      if(pass && ((index1==playernum-1) || (index2==playernum-1))) Error("game format: invalid player");
		      if(!(getline(in, s))) Error("game format: result string"); //***COME BACK: could do a comment check because include comments in result
		      
		      matrix[temp_r*games_per_pair+gameNum][index1].opponent = index2;
		      matrix[temp_r*games_per_pair+gameNum][index2].opponent = index1;
		      matrix[temp_r*games_per_pair+gameNum][index1].seed = matrix[temp_r*games_per_pair+gameNum][index2].seed = game_seed;
		      //can't be a dummy if have 2 real players:
		      matrix[temp_r*games_per_pair+gameNum][index1].dummy = matrix[temp_r*games_per_pair+gameNum][index2].dummy = false;
		      matrix[temp_r*games_per_pair+gameNum][index1].result = matrix[temp_r*games_per_pair+gameNum][index2].result = s;
		      matrix[temp_r*games_per_pair+gameNum][index1].valid = matrix[temp_r*games_per_pair+gameNum][index2].valid = true;
		      roundflag = false;
		      gameflag = false;
	
		     
		    }
		}
	      else
		getline(in, s);
	    }
	  else
	    getline(in, s);
	}
      //    if(gameNum >= games_per_pair-1)
	++roundNum; //we want to run the next round(if no results are at the bottom of the file, want to run round 0 so don't increment)don't need this
                    //here, roundnum is the last round which contains a hole or the following round which needs to be completely played
    }
  //Print out the schedule with added state
  displayMatrix(matrix, playernum, games_per_pair, single);
  int col; //column index
  if(opt_s && roundNum <= playernum-1)
    {     
      in.close();
      ofstream out(argv[argi], ios::app);
      if(!out.is_open()) Error("unable to open the tournament file for writing");
      out << endl;
      
      //check for holes & play them all
      FOR(i, roundNum)
	{
	  FOR(j, games_per_pair)
	    {
	      if((matrix[i*games_per_pair+j][pair[0]].opponent == pair[1]) && (matrix[i*games_per_pair+j][pair[0]].valid == false))
		{
		  play(matrix, pair[0], i, games_per_pair, j);
		  out << "r" << i+1 << " g" << j+1 << " ";	
		  
		  if(!single && j&1) 
		    {
		      if(pair[1]>pair[0])
			col = pair[1];
		      else
			col = pair[0];
		    }
		  else
		    {
		      if(pair[1]<pair[0])
			col = pair[1];
		      else
			col = pair[0];
		    } 
		  out << matrix[i*games_per_pair+j][col].seed << " ";
		  out << players[col]->name << "\t" << players[matrix[i*games_per_pair+j][col].opponent]->name << "\t";	 
		  out << matrix[i*games_per_pair+j][col].result << endl;
		  matrix[i*games_per_pair+j][col].seen = matrix[i*games_per_pair+j][matrix[i*games_per_pair+j][col].opponent].seen = true;
		}
	    }
	  cout << "CHECK FOR HOLES!!!!!!!!!!!!!!!!!!!!\n";
	  
	}
      out.close();
    }   
  else if(roundNum <= playernum-1) //added '=' here because even if all rounds have data, have to play the first incomplete round
    {
      roundflag = false;
      in.close();
      ofstream out(argv[argi], ios::app);
      if(!out.is_open()) Error("unable to open the tournament file for writing");
      out << endl;

      //Play the next round (***COME BACK: regardless of if the previous round's games are done!!!) - only if previous rounds are complete
      FOR(roundNum,playernum-1)          //UNDERMINE ROUNDNUM'S VALUE HERE -- DECIDE IF THIS IS OK?!?
	{
     	  FOR(gameNum, games_per_pair)
	    {
	      cout << "Run round: " << roundNum << "\t game: " << gameNum << endl;
	      FOR(j, playernum)
		{
		  if(!(matrix[roundNum*games_per_pair+gameNum][j].valid))
		    {
		      if(!(matrix[roundNum*games_per_pair+gameNum][j].dummy))
			{
			  play(matrix, j, roundNum, games_per_pair, gameNum);
			 
			  roundflag = true;
			}
		      matrix[roundNum*games_per_pair+gameNum][j].valid = true;
		      matrix[roundNum*games_per_pair+gameNum][matrix[roundNum*games_per_pair+gameNum][j].opponent].valid = true;
		      
		      if(!(matrix[roundNum*games_per_pair+gameNum][j].seen) && !(matrix[roundNum*games_per_pair+gameNum][j].dummy)) //***COME BACK:what about checking for valid?????
			{
			  out << "r" << roundNum+1 << " g" << gameNum+1 << " ";
			  if(!single && gameNum&1) col = matrix[roundNum*games_per_pair+gameNum][j].opponent;
			  else col = j;
			  out << matrix[roundNum*games_per_pair+gameNum][col].seed << " ";
			  out << players[col]->name << "\t" << players[matrix[roundNum*games_per_pair+gameNum][col].opponent]->name << "\t";
			  out << matrix[roundNum*games_per_pair+gameNum][col].result << endl;
			  matrix[roundNum*games_per_pair+gameNum][col].seen = matrix[roundNum*games_per_pair+gameNum][matrix[roundNum*games_per_pair+gameNum][col].opponent].seen = true;
			}
		      
		    }
		}
	    }
	  if(roundflag)
	    break;
	}
    } 

  cout << "done round " << roundNum <<" !\n";
  //Print out updated schedule and tournament state
  displayMatrix(matrix, playernum, games_per_pair, single);

  
//SSH notes: 
/*  FILE *fpREAD =popen("ssh krysta@ug01.cs.ualberta.ca ./WORK/SERVER", "r");                           */
/*  get a warning with popen: Warning: No xauth data; using fake authentication data for X11 forwarding.*/
/*  //what's the difference bt/n -f and & for the ssh command -- will this screw stuff up?              */
/*  FILE *fpREAD =popen("ssh krysta@ug01.cs.ualberta.ca ./WORK/PLAYER >> server_results 2>&1 &");       */
		      
  //delete the matrix:
  FOR(i, (playernum-1)*games_per_pair)
    delete [] matrix[i];
  delete [] matrix;
  //delete players map data; 
  // ***COME BACK: VALGRIND, STILL MAY LOSE MEM SOMEWHERE(when no ip file)(163 allocs 155 frees)!!!!!
  FOR(i, playernum)
    delete players[i];
  
  return 0;
}
