// $Id: Octree.C,v 1.16 2005/07/27 07:15:11 orts_furtak Exp $
 
// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "GfxObject.H"
#include "Octree.H"
#include "GameObj.H"
#include "Profiler.H"

using namespace std;
REGISTER_TYPEOF(240, Vector<Sector*>::iterator);

#define MEASURE_CYCLES 0

#if MEASURE_CYCLES
long long total_to_be_drawn = 0;
#endif

Octree::Octree(const Coor3 &orig, const Coor3 &dim, sint4 levels)
{
  world_orig = orig;
  world_dim  = dim;
  sec_levels = levels;
  culling_mode = OCTREE;

  if (dim.x <= 0 || dim.y <= 0 || dim.z <= 0) ERR("negative dimension");

  FORS (i, sec_levels) {
    sint4 n = 1;
    FORS (j, i) n *= 8;
    sectors.push_back(new Sector[n]);
    sectors_n.push_back(n);
    cout << i << " " << n << " sectors" << endl;
  }
}


Octree::~Octree()
{
  FORS (i, sec_levels) delete [] sectors[i];
}


sint4 Octree::used() const {
  sint4 n = sectors_n[sec_levels-1];
  sint4 u = 0;
  FORS (i, n) u += sectors[sec_levels-1][i].used;
  return u;
}


// computes equidistant index range in one direction of line (x1,x2)
// in (xmin,xmax) which is covered by nsec segments

void Octree::index_range(real4 x1, real4 x2, real4 xmin, real4 xmax, sint4 nsec,
			sint4 &xi_min, sint4 &xi_max)
{
  if (x1 > x2) swap(x1, x2);

  if (x1 < xmin) x1 = xmin;
  if (x2 > xmax) x2 = xmax;

  real4 inv = nsec/(xmax-xmin);
  
  xi_min = sint4(floor((x1-xmin)*inv));
  if (xi_min < 0) xi_min = 0; else if (xi_min >= nsec) xi_min = nsec-1;

  xi_max = sint4(((x2-xmin)*inv));
  if (xi_max < 0) xi_max = 0; else if (xi_max >= nsec) xi_max = nsec-1;

  //  cout << "index: " << x1 << " " << x2 << " " << xmin << " " << xmax << " n= " << nsec
  //       << " -> " << xi_min << " " << xi_max << endl;
}

void Octree::update(GfxObject *gob)
{
  if (culling_mode == OCTREE) {
    remove(gob);
    add(gob);
  }
}

void Octree::add(GfxObject *gob)
{
  ++add_count;

  if (culling_mode != OCTREE) {
    all_objs.push_back(gob);
    return;
  }


  // get axis aligned bounding box
  const Vec3<real4> &o = gob->aabb_o;
  const Vec3<real4> &d = gob->aabb_d;

  // compute sector ranges in three dimensions

  sint4 shift = sec_levels-1;
  sint4 ndim = 1 << shift;

  sint4 xi_min, xi_max;
  index_range(o.x, o.x+d.x, world_orig.x, world_orig.x+world_dim.x, ndim, xi_min, xi_max);

  sint4 yi_min, yi_max;
  index_range(o.y, o.y+d.y, world_orig.y, world_orig.y+world_dim.y, ndim, yi_min, yi_max);

  sint4 zi_min, zi_max;
  index_range(o.z, o.z+d.z, world_orig.z, world_orig.z+world_dim.z, ndim, zi_min, zi_max);

  // put pointer to object into all intersecting cubes
  for (sint4 x=xi_min; x <= xi_max; ++x) {
    for (sint4 y=yi_min; y <= yi_max; ++y) {
      for (sint4 z=zi_min; z <= zi_max; ++z) {

	assert(ndim == (1 << shift));
        
	sint4 ind = (((z << shift) + y) << shift) + x;
	assert(ind >= 0 && ind < sectors_n[sec_levels-1]);
	sectors[sec_levels-1][ind].used = true;
	sectors[sec_levels-1][ind].objs.push_back(gob);
	gob->sectors.push_back(&sectors[sec_levels-1][ind]);
        
	// mark path to root as being used
	sint4 x2 = x, y2 = y, z2 = z, sh2 = shift;;
	for (sint4 l = sec_levels-2; l >= 0; --l) {
	  --sh2;
	  x2 >>= 1; y2 >>= 1; z2 >>= 1;
	  sint4 ind2 = (((z2 << sh2) + y2) << sh2) + x2;
	  assert(ind2 >= 0 && ind2 < sectors_n[l]);
	  if (sectors[l][ind2].used) break;
	  sectors[l][ind2].used = true;
	}
      }
    }
  }
}

void Octree::remove(GfxObject *gob)
{
  ++remove_count;

  if (culling_mode != OCTREE) {
    FORALL_REV_B (all_objs, i) if ((*i) == gob) break;
    if (i == all_objs.rend()) return; //ERR("gfxobj not in octree!");
    assert(&*i == &*(i.base()-1));
    all_objs.erase(i.base()-1);
    return;
  }

  FORALL (gob->sectors, s) {
    if (!(*s)->used) ERR("sector unused!");

    // find gob in sector objs
    FORALL_REV_B ((*s)->objs, i) if ((*i) == gob) break;
    if (i == (*s)->objs.rend()) ERR("gfxobj not in octree!");

    assert(&*i == &*(i.base()-1));
    
    (*s)->objs.erase(i.base()-1);

    // FLAG: this got called when running - is it an actual error?

#ifndef NDEBUG
    //    FORALL ((*s)->objs, i) if ((*i) == gob) ERR("WHAT");
#endif
    
    if ((*s)->objs.empty()) {
      (*s)->used = false;
      // fixme: follow path to root and check whether sectors become unused
      // in the meantime, used == true in interior nodes does not mean there
      // are used objects in the subtree ...
      
      // sint4 ind = &(*s) - sectors[sec_levels-1]; ...
    }
  }
  gob->sectors.clear();
}


// o,d:      current cube (origin, dimension)
// level:    at which octree level are we? starts with 0
// xi,yi,zi: current cube indexes

void Octree::draw_scene(const Coor3 &o, const Coor3 &d, sint4 level,
                        sint4 xi, sint4 yi, sint4 zi,
	                Vector<GfxObject*> &to_be_drawn)
{
  //  cout << level << " " << xi << " " << yi << " " << zi << endl;

  sint4 ind = (((zi << level) + yi) << level) + xi;

  assert(ind >= 0 && ind < sectors_n[level]);

  if (!sectors[level][ind].used) return;

  if (!fr.aabox_intersects(o, d)) return;

  if (level == sec_levels-1) {

    PROFILE("Octree - draw object");
    // leaf: draw all objects

    FORALL (sectors[level][ind].objs, go) {
      GfxObject *gob = *go;

#if MEASURE_CYCLES
      static long long c = 0, n = 0;
#endif
      bool res = gob->drawn != drawn_magic;

#if MEASURE_CYCLES      
      n++;
      // average cycles (minimum ~ 48), number of drawn checks per object drawn
      if ((n & 65535) == 65535)
	cout << double(n)/total_to_be_drawn << endl;
#endif
      
      if (res) {
	
	gob->drawn = drawn_magic;
	if (!gob->dead) to_be_drawn.push_back(gob);
      }
    }

    return;
  }

  // draw children

  // size of children boxes
  Coor3 h = d;
  h.x *= 0.5;
  h.y *= 0.5;
  h.z *= 0.5;

  // one level up
  ++level;
  xi *= 2;
  yi *= 2;
  zi *= 2;
  
  {
    Coor3 n = o;
    draw_scene(n, h, level, xi, yi, zi, to_be_drawn);
  }
  {
    Coor3 n = o; n.x += h.x;
    draw_scene(n, h, level, xi+1, yi, zi, to_be_drawn);
  }
  {
    Coor3 n = o; n.y += h.y;
    draw_scene(n, h, level, xi, yi+1, zi, to_be_drawn);
  }
  {
    Coor3 n = o; n.x += h.x; n.y += h.y;
    draw_scene(n, h, level, xi+1, yi+1, zi, to_be_drawn);
  }
  {
    Coor3 n = o; n.z += h.z;
    draw_scene(n, h, level, xi, yi, zi+1, to_be_drawn);
  }
  {
    Coor3 n = o; n.z += h.z; n.x += h.x;
    draw_scene(n, h, level, xi+1, yi, zi+1, to_be_drawn);
  }
  {
    Coor3 n = o; n.z += h.z; n.y += h.y;
    draw_scene(n, h, level, xi, yi+1, zi+1, to_be_drawn);
  }
  {
    Coor3 n = o; n.z += h.z; n.x += h.x; n.y += h.y;
    draw_scene(n, h, level, xi+1, yi+1, zi+1, to_be_drawn);
  }
}


void Octree::draw_culled(Vector<GfxObject*> &to_be_drawn)
{
  Mat4<real4> m;
  glGetModelview(m);
  m.invert();
  
  Vec4<real4> p(0, 0,  0, 1); p = m*p; // camera position
  Vec4<real4> r(0, 0, -1, 0); r = m*r; // normal vector
  
  real8 c2 = 1.0 / (r.x*r.x + r.y*r.y + r.z*r.z);
  
  FORALL (all_objs, go) {
    GfxObject &gob = **go;
    
    if (gob.dead) continue;

    const Vec3<real4> &mid = gob.pose.center;
    
    // dot product of camera normal with vector to the object
    real4 c1 = r.x * (mid.x-p.x) + r.y * (mid.y-p.y) + r.z * (mid.z-p.z);
    real4 b = c1 * c2;
    real4 pbx = p.x + b*r.x;
    real4 pby = p.y + b*r.y;
    real4 pbz = p.z + b*r.z;
    
    real4 d = square(pbx - mid.x) + square(pby - mid.y) + square(pbz - mid.z);
    //    cout << "d = " << d << endl;
    
    if (d < 580.0) { // fixme: magic constant?
      to_be_drawn.push_back(&gob);
    }
  }
  
  return;
}

void Octree::draw_all(Vector<GfxObject*> &to_be_drawn)
{
  to_be_drawn = all_objs;
}

// draw all objects which intersect with frustum

void Octree::draw(Vector<GfxObject*> &to_be_drawn)
{
  to_be_drawn.clear();

  if (culling_mode == OCTREE) {
    fr.compute();
    drawn_magic++;
    
    if (drawn_magic >= 1000000) {
      
      sint4 n = sectors_n[sec_levels-1];
      FORS (i, n) {
        FORALL (sectors[sec_levels-1][i].objs, j) (*j)->drawn = 0;
      }
      
      drawn_magic = 1;
    }
    
    draw_scene(world_orig, world_dim, 0, 0, 0, 0, to_be_drawn);

#if MEASURE_CYCLES    
    total_to_be_drawn += to_be_drawn.size();
#endif
    
  } else if (culling_mode == CONE) {

    draw_culled(to_be_drawn);

  } else if (culling_mode == NO_CULLING) {

    draw_all(to_be_drawn);

  }
}

