#ifndef ScriptGame_H
#define ScriptGame_H

/** @file ScriptGame.H
    Object action scripting language.

    $Id: ScriptGame.H,v 1.10 2006/07/30 16:18:47 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/ScriptGame.H,v $
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Script.H"

class ScriptObj;
class GameObjFactory;
class SimpleActions;
class ObjAction;
class PlayerInfo;
class Game;
class ScriptParser;

//=============================================================================
//=== object manipulators =====================================================
//=============================================================================

/** Dereferences a list of subcomponent names, stores the result. */
class GOB_Component : public ScriptNode {
  stype obj;         ///< Set to the final object dereferenced by name.
  stype *gob;        ///< The object whose component(s) are being accessed.
  Vector<std::string> comp_names; ///< A sequence of component names to dereference.
  ScriptNode *first; ///< For global objects, locate the base object first.
public:
  GOB_Component(stype *g, Vector<std::string> &comp);
  ~GOB_Component() { if (first) delete first; }
  stype eval();
  friend class ScriptParser;
};

//=============================================================================
//=== atomic values ===========================================================
//=============================================================================

/** Looks through an object's components and attributes and returns the value. */
class GOB_AttrNode : public VariableNode {
  std::string attr;
  stype *gob;
  /** A list of component dereferencing prior to the final element: i.e. base.comp1.comp2...compN.z */
  GOB_Component *par;
public:
  GOB_AttrNode(const std::string &a, stype *g, GOB_Component *p = 0)
    : attr(a), gob(g), par(p)
  {
    //    if (!!gob) ScriptObj::check_ptr(gob->to_obj());
  }
  stype eval();
  stype* addr();
};

//-----------------------------------------------------------------------------

class GOB_AddrNode : public VariableNode {
  stype *gob;
  GOB_Component *par;
public:

  GOB_AddrNode(stype *g, GOB_Component *p = 0) : gob(g), par(p) {
    //    ScriptObj::check_ptr(gob->to_obj());
  }

  stype eval();
  stype* addr() { if (par) par->eval(); return gob; }
  friend class ScriptParser;
};

//-----------------------------------------------------------------------------

class BP_AttrNode : public VariableNode {
  Vector<std::string> attrs;
  GameObjFactory *gof;
  stype *ptr;
  void get_ptr();
public:
  BP_AttrNode(const Vector<std::string> &a, GameObjFactory *f);
  stype eval();
  stype* addr();
};

//=============================================================================
//=== assignment operators ====================================================
//=============================================================================

//=============================================================================
//=== functions ===============================================================
//=============================================================================

class LostSight : public FunctionNode {
protected:
  VariableNode *gob;
  ScriptNode *player;
public:
  LostSight() { gob = 0; player = 0; }
  ~LostSight() { delete gob; gob = 0; delete player; player = 0; }
  ScriptNode* parse(ScriptParser &parser);
  stype eval();
};
//-----------------------------------------------------------------------------
class NumIntersect : public FunctionNode {
protected:
  VariableNode *gob;
public:
  NumIntersect() { gob = 0; }
  ~NumIntersect() { delete gob; gob = 0; }
  ScriptNode* parse(ScriptParser &parser);
  stype eval();
};
//-----------------------------------------------------------------------------
class BP_Code : public FunctionNode {
  std::string bp_name;
public:
  BP_Code() {}
  ScriptNode* parse(ScriptParser &parser);
  stype eval();
};
//-----------------------------------------------------------------------------
class GlobalObj : public ScriptNode {
  stype obj;
  std::string name;
  PlayerInfo *pi;
  //  ScriptNode *next; // next command after dereferencing
public:
  GlobalObj(std::string n, PlayerInfo *p) : name(n), pi(p) { }
  ~GlobalObj() {}
  stype eval();
  friend class ScriptParser;
};
//-----------------------------------------------------------------------------
class SharedObj : public ScriptNode {
  stype obj;
  std::string name;
public:
  SharedObj(std::string n) : name(n) { }
  ~SharedObj() {}
  stype eval();
  friend class ScriptParser;
};

//=============================================================================
//=============================================================================

class ForAllObj : public ScriptNode {
protected:
  stype *gob;
  ScriptNode *expression;
  Vector<ScriptObj::Ptr> *objs; // objects to check
public:
  uint4 PC;

  ForAllObj(stype *g, Vector<ScriptObj::Ptr> *objs);
  ~ForAllObj() { if (expression) delete expression; expression = 0; }
  void set_body(ScriptNode *body) { expression = body; }
  stype eval();
};
//-----------------------------------------------------------------------------
class ForAllType : public ScriptNode {
protected:
  stype *gob;
  ScriptNode *expression;
  Vector<ScriptObj::Ptr> *objs; // objects to check
  std::set<std::string> bp_types;
public:
  uint4 PC;

  ForAllType(stype *g, Vector<ScriptObj::Ptr> *objs, std::set<std::string> &bp_types);
  ~ForAllType() { if (expression) delete expression; expression = 0; }
  void set_body(ScriptNode *body) { expression = body; }
  stype eval();
};
//-----------------------------------------------------------------------------
class ForAllIntersect : public ScriptNode {
protected:
  stype *gob; // intersects area, assigned to gob
  VariableNode *area; // object to check intersections with
  ScriptNode *expression;
  Game *game;
public:
  uint4 PC;

  ForAllIntersect(stype *o, VariableNode *a, Game *g) :
    gob(o), area(a), expression(0), game(g) {
    ScriptObj::check_ptr(gob->to_obj());
  }
  ~ForAllIntersect() { if (expression) delete expression; expression = 0; }
  void set_body(ScriptNode *body) { expression = body; }
  stype eval();
};

//=============================================================================

#endif
