#include "OgreMD3Model.H"

/*==========================================================================================
load(string &modelfile)
-This class takes the locatrion of the modelfile assumes its a MD3 and does everything required
to load it.  Everything is then stored into the private variablese inside this class.
-After calling load a call to createMD3Object(SceneManager* mgr, string &modelName) must also
be called which will load the object into ogre useable code. See createMD3object for more info.
==========================================================================================*/
sint4 OgreMD3Model::load(const std::string &modelfile){

  // standard md3 file
  FILE *f = fopen(modelfile.c_str(), "rb");  //attempt to open file.
  if (!f) {  ERR2("couldn't open file", modelfile.c_str());  return 0;  }

  MD3Header header;   //load MD3 Header
  if (!load_header(f, header)){ ERR2("couldn't loader header for ", modelfile.c_str()); return 0; }
  nvert = 0;
  npoly = 0;

  MD3Surface *surfaces = new MD3Surface[header.numSurfaces];  //allocate memory for surfaces.

  for(int i=0; i < header.numSurfaces; i++){  //read the surfaces from file and store them.
    fseek(f, header.offsetSurfaces, SEEK_SET);
    if (fread(&surfaces[i], sizeof(MD3Surface), 1, f) != 1) {
      fclose(f); 
      delete [] surfaces;
      return 0;
    }

    if (surfaces[i].id != MD3_ID) {  //make sure they are md3 files.
      fclose(f); 
      delete [] surfaces;
      return 0;
    }
    nvert += surfaces[i].numVerts;
    npoly += surfaces[i].numTriangles;
    surfaces[i].offsetVerts += header.offsetSurfaces;
    surfaces[i].offsetST += header.offsetSurfaces;
    surfaces[i].offsetTriangles += header.offsetSurfaces;
    header.offsetSurfaces += surfaces[i].offsetEnd;
  };

  if (!load_tags(f, header)) {  fclose(f);  return 0; }  //load tags.
  bbox.resize(header.numFrames);
  vert    = new Vec3<real4>[nvert*header.numFrames];
  vnorm   = new Vec3<real4>[nvert*header.numFrames];
  tvert   = new real4[nvert*2];
  memset(tvert, 0, sizeof(tvert[0])*nvert*2);

  poly = new Poly[npoly];

  int vert_offset = 0;
  int poly_offset = 0;

  sint4 *buffer = new sint4[3*npoly];

  for(int s=0; s < header.numSurfaces; s++){
    fseek(f, surfaces[s].offsetST, SEEK_SET);  // read texture coords
    
    if (fread(tvert + 2*vert_offset, sizeof(real4)*2*surfaces[s].numVerts, 1, f) != 1) {
      delete [] buffer;      
      delete [] surfaces;
      return 0;
    }  // read vertices
   
    MD3Vertex tmp;

    fseek(f, surfaces[s].offsetVerts, SEEK_SET); // fixme: check if OK

    for(int j=0; j < header.numFrames; j++){
      for(int i=0; i < surfaces[s].numVerts; i++){
        if (fread(&tmp, sizeof(tmp), 1, f) != 1) {
	  delete [] buffer;	  
          delete [] surfaces;
          fclose(f); return 0;
        }

        sint4 ind = nvert*j + i + vert_offset;
	assert(ind < (sint4)(nvert*header.numFrames));
        vert[ind] = Vec3<real4>(tmp.x, tmp.z, -tmp.y).rotate(Vec3<real4>(0, 1, 0), M_PI/2)*(1.0/1000.0);
        //calculating normals.
        real4 lat = (tmp.normal & 255) * (2 * M_PI)/255.0;
        real4 lng = ((tmp.normal >> 8) & 255) * (2 * M_PI)/255.0;
        vnorm[ind].x = cos(lat) * sin(lng);
        vnorm[ind].y = sin(lat) * sin(lng);
        vnorm[ind].z = cos(lng);
      }
    }
    
    fseek(f, surfaces[s].offsetTriangles, SEEK_SET);  // read triangles

    if (fread(buffer, sizeof(buffer[0])*3*surfaces[s].numTriangles, 1, f) != 1) {
      delete [] buffer;
      delete [] surfaces;
      fclose(f); return 0;
    }
    for(int i=0; i < surfaces[s].numTriangles; i++){
      for(int j=0; j<3; j++){
        //little2machine(buffer[i*3+j]);
        poly[i + poly_offset].vert[j] = buffer[i*3+j] + vert_offset; 
      }
    }
    poly_offset += surfaces[s].numTriangles;
    vert_offset += surfaces[s].numVerts;
  }
  fclose(f);
  delete [] buffer; buffer = 0;
  delete [] surfaces; surfaces = 0;

  // try to find a cfg file with animation names
  std::string fname = modelfile.substr(0, modelfile.length()-4) + ".cfg";
  load_cfg(fname);

  return 1;
}

/*==========================================================================================
createMD3Object(SceneManager* mgr, string &modelName)
This class takes the loaded MD3 object private variables vert, nvert etc.. and loads them into
Ogre useable code.
TODO: Bounding boxes are not yet properly working.  
TODO: (ANIMATION)  explained.  Due to time reasons was unable to get animation working properly.
Basically right now I disabled it because its only like 1/4 done.  There is a Ogre class called
VertexMorhKeyFrame I believe this is the key to getting the animation working. Ogre animation is 
suppose to use skeleton but we dont have models for this we only have keyframe(md3) which is really
old and basically never sued. Documentation is VERY limited atm so its going to take some time to 
figure this out. All the keyframes can be accessed by just looking at the vert[i + keyframeNum * nvert] 
and vnorm[i + keyframeNum * nvert] and what each keyframe does can be found in the .cfg files which are 
loaded and stored in teh name2frames std::pair variable.
==========================================================================================*/
void OgreMD3Model::createMD3Object(Ogre::SceneManager* mgr, const std::string &modelName, const int scale){
  std::pair<sint2, sint2> temp = name2frames["idle"];
  unsigned int s1 = temp.first;
  unsigned int s2 = temp.second;

  Ogre::MeshPtr msh = Ogre::MeshManager::getSingleton().createManual(modelName, "General");
  Ogre::SubMesh *sub = msh->createSubMesh();
    //create normal / vertex mesh.
  float vertices[nvert*6];
  unsigned int x=0;
  int gap = nvert * 0;
  for( unsigned int i=0; i < nvert; i++){
    vertices[x++] = vnorm[i+gap].x;         //normal 1
    vertices[x++] = vnorm[i+gap].y;         //normal 2
    vertices[x++] = vnorm[i+gap].z;         //normal 3
    vertices[x++] = vert[i+gap].x*scale;    //vertex 1    
    vertices[x++] = vert[i+gap].y*scale;    //vertex 2
    vertices[x++] = vert[i+gap].z*scale;    //vertex 3
  }

  //create texture mesh. The same texture mesh is used for all Keyframes.
  float textures[2*nvert];
  unsigned int x1=0;
  for( unsigned int i=0; i < nvert; i++){
    textures[x1++] = tvert[i*2];         //texCoord 1
    textures[x1++] = tvert[i*2+1];       //texCoord 2
  }

  //create face index's.
  const size_t ibufCount = npoly*3;
  unsigned short faces[ibufCount];
  unsigned int counter = 0;
  for( unsigned int i=0; i < npoly; i++ ){  //create array of Faces.
    faces[counter++] = poly[i].vert[2];
    faces[counter++] = poly[i].vert[1];
    faces[counter++] = poly[i].vert[0];
  }

  msh->sharedVertexData = new Ogre::VertexData();
  msh->sharedVertexData->vertexCount = nvert;

  Ogre::VertexDeclaration* decl = msh->sharedVertexData->vertexDeclaration;
  size_t offset = 0;
  //1st buffer.
  decl->addElement(0, offset, Ogre::VET_FLOAT3, Ogre::VES_NORMAL);
  offset += Ogre::VertexElement::getTypeSize(Ogre::VET_FLOAT3);
  decl->addElement(0, offset, Ogre::VET_FLOAT3, Ogre::VES_POSITION);
  offset += Ogre::VertexElement::getTypeSize(Ogre::VET_FLOAT3);
  //Allocate vertex buffer of the requested number of vertices (vertexCount) 
  //and bytes per vertex (offset)
  Ogre::HardwareVertexBufferSharedPtr vbuf = Ogre::HardwareBufferManager::getSingleton().createVertexBuffer(
    offset, msh->sharedVertexData->vertexCount, Ogre::HardwareBuffer::HBU_STATIC_WRITE_ONLY);
  //Upload the vertex data to the card
  vbuf->writeData(0, vbuf->getSizeInBytes(), vertices, true);

  //Set vertex buffer binding so buffer 0 is bound to our vertex buffer
  Ogre::VertexBufferBinding* bind = msh->sharedVertexData->vertexBufferBinding; 
  bind->setBinding(0, vbuf);
  
  //2nd buffer.
  offset = 0;
  decl->addElement(1, offset, Ogre::VET_FLOAT2, Ogre::VES_TEXTURE_COORDINATES);
  offset += Ogre::VertexElement::getTypeSize(Ogre::VET_FLOAT2);
  vbuf = Ogre::HardwareBufferManager::getSingleton().createVertexBuffer(offset, msh->sharedVertexData->vertexCount,
								Ogre::HardwareBuffer::HBU_STATIC_WRITE_ONLY);
  vbuf->writeData(0, vbuf->getSizeInBytes(), textures, true);
  bind->setBinding(1, vbuf);
  /// Allocate index buffer of the requested number of vertices (ibufCount) 
  Ogre::HardwareIndexBufferSharedPtr ibuf = Ogre::HardwareBufferManager::getSingleton().
    createIndexBuffer(  Ogre::HardwareIndexBuffer::IT_16BIT, ibufCount, Ogre::HardwareBuffer::HBU_STATIC_WRITE_ONLY);
  /// Upload the index data to the card
  ibuf->writeData(0, ibuf->getSizeInBytes(), faces, true);
  /// Set parameters of the submesh
  sub->useSharedVertices = true;
  sub->indexData->indexBuffer = ibuf;
  sub->indexData->indexCount = ibufCount;
  sub->indexData->indexStart = 0;

  /// Set bounding information (for culling)
  msh->_setBounds(Ogre::AxisAlignedBox(-25,-25,-25,25,25,25));
  msh->_setBoundingSphereRadius(Ogre::Math::Sqrt(3*25*25));

  /// Notify Mesh object that it has been loaded
  msh->load();

  Ogre::Entity *entity;
  entity = mgr->createEntity(modelName, modelName);
  Ogre::SceneNode *objNode;
  objNode = mgr->getRootSceneNode()->createChildSceneNode();

  entity->setMaterialName(modelName);

  objNode->attachObject( entity );
	
  delete [] vert;  //this is all now stored in HardwareVertexBufferSharedPtr's delete.
  delete [] tvert;
  delete [] poly;  //stored in HardwareIndexBuffer
  delete [] vnorm;
}

//Loads the cfg file which holds information on how the animation works
//eg.   (  0	30	30	15		// turn  ) -> scope.cfg
void OgreMD3Model::load_cfg(const std::string &fname){
  std::ifstream file;
  file.open(fname.c_str());
  if (!file.is_open()){  return; }
  
  while (file.good()) {
    char line[256];
    file.getline(line, 256);
    std::string ln = line;
    for(unsigned int i=0; i < ln.size(); i++){
      if (ln[i] == '\r' || ln[i] == '\n') {
        ln = ln.substr(0, i);
        break;
      }
    }
    if (ln == "" || ln.substr(0, 2) == "//") continue;
    std::istringstream ss(ln);
    sint4 first, num, looping, fps;
    ss >> first;
    ss >> num;
    ss >> looping;
    ss >> fps;
    
    std::string temp;
    ss >> temp;
    if (temp != "//") REM("expecting animation name");
    ss >> temp;
    name2frames[temp] = std::pair<sint2, sint2>(first, first+num-1);
  }
  file.close();
}

//loads teh header file for information on teh rest of the file including offsets etc.
bool OgreMD3Model::load_header(FILE* &f, MD3Header &header){
  if (fread(&header, sizeof(header), 1, f) != 1) {
    fclose(f); 
    return false;
  }
  if (header.id != MD3_ID) {
    fclose(f);    
    return false;
  }
  return true;
}

//loads teh tags which is used for weapons etc  Bascially its another object which moves
//with the main object.
bool OgreMD3Model::load_tags(FILE* &f, const MD3Header &header){
  if (header.numTags) {
    MD3Tag *tags = new MD3Tag[header.numTags * header.numFrames];
    fseek(f, header.offsetTags, SEEK_SET);
    if (fread(tags, sizeof(MD3Tag) * header.numTags * header.numFrames, 1, f) != 1) {
      delete [] tags;
      return false;
    }
    //FORS (i, header.numTags * header.numFrames) tags[i].little2machine();

    all_tags = new ModelTag[header.numFrames*header.numTags];
    tag.resize(header.numTags);
    tag_name.resize(header.numTags);
    FORS (i, header.numTags) {
      tag[i] = &all_tags[header.numFrames*i];
      tag_name[i] = reinterpret_cast<char*>(tags[i].name);
      if (tag_name[i].size() > 50) tag_name[i] = tag_name[i].substr(0,50);
    }

    FORS (i, header.numFrames) {
      FORS (j, header.numTags) {
        sint4 ind = j + i*header.numTags;
        real4 (*a)[3] = tags[ind].axis;
        // adjust the axis for tags to match the default model transformation
        Vec3<real4> ax(a[1][0], a[1][1], a[1][2]);
        Vec3<real4> ay(a[2][0], a[2][1], a[2][2]);
        Vec3<real4> az(a[0][0], a[0][1], a[0][2]);

        Vec3<real4> of(tags[ind].origin[0],
                       tags[ind].origin[1],
                       tags[ind].origin[2]);

        // transform md3 coordinates into our system
        // tag coordinates are 1/64 of the vertex scale
        // 1.0/15.625 = magic constant based on 1000 scaling factor below
        // (1000 / 64 = 15.624)

        ax = Vec3<real4>(ax.x, ax.z, -ax.y).rotate(Vec3<real4>(0, 1, 0), M_PI/2);
        ay = Vec3<real4>(ay.x, ay.z, -ay.y).rotate(Vec3<real4>(0, 1, 0), M_PI/2);
        az = Vec3<real4>(az.x, az.z, -az.y).rotate(Vec3<real4>(0, 1, 0), M_PI/2);
        of = Vec3<real4>(of.x, of.z, -of.y).rotate(Vec3<real4>(0, 1, 0), M_PI/2) * (1.0/15.625);

        Vec4<real4> axis_x(ax.x, ax.y, ax.z, 0);
        Vec4<real4> axis_y(ay.x, ay.y, ay.z, 0);
        Vec4<real4> axis_z(az.x, az.y, az.z, 0);
        Vec4<real4> foo   (of.x, of.y, of.z, 1); // position offset

        //        axis_x.normalize();
        //        axis_y.normalize();
        //        axis_z.normalize();

        tag[j][i].axis = Mat4<real4>(axis_x, axis_y, axis_z, foo);				
      }
    }
    delete [] tags;
  }
  return true;
}
