(defvar bp-mode-hook nil)
(defvar bp-mode-map nil
  "Keymap for BP major mode.")

(if bp-mode-map nil
  (setq bp-mode-map (make-keymap)))

(setq auto-mode-alist
	  (append
	   '(("\\.bp\\'" . bp-mode))
	   auto-mode-alist))


(defconst bp-font-lock-keywords-1
  (list
   (append (list (concat "\\<"
			 (regexp-opt '("blueprint" "class" "end" "intersect" "enable" "disable" "cancel" "create" "create_id" "update" "position" "kill" "move" "notify_attack" "can_see" "distance" "visible_by" "lost_sight" "rand" "alive" "dead" "type" "typeof") t)
			 "\\>"))
	   'font-lock-builtin-face)
   '("#.*$" . font-lock-comment-face)
   )
  "BP included functions.")

(defconst bp-font-lock-keywords-2
  (append bp-font-lock-keywords-1
	  (list
	   (append (list (concat "\\<"
				 (regexp-opt '("forall" "if" "else" "while" "return" "break" "this" "global") t)
				 "\\>"))
		   'font-lock-keyword-face)))
  "Additional Keywords to highlight in BP mode.")

(defconst bp-font-lock-keywords-3
  (append bp-font-lock-keywords-2
	  (list
	   (append (list (concat "\\<"
				 (regexp-opt '("npc" "me" "tick") t)
				 "\\>"))
		   'font-lock-constant-face)))
  "Additional Keywords to highlight in BP mode.")

(defconst bp-font-lock-keywords-4
  (append bp-font-lock-keywords-3
	  (list
	   (append (list (concat "\\<"
				 (regexp-opt '("action" "constructor" "destructor") t)
				 "\\>"))
		   'font-lock-function-name-face)))
  "Actions in BP mode.")

(defconst bp-font-lock-keywords-5
  (append bp-font-lock-keywords-4
	  (list
	   (append (list (concat "\\<"
				 (regexp-opt '("int" "var" "gob" "obj" "set" "setf" "const" "hidden" "internal" "has" "is" "BP") t)
				 "\\>"))
		   'font-lock-type-face)))
  "Variables in BP mode.")


(defvar bp-font-lock-keywords bp-font-lock-keywords-5
  "Default highlighting expressions for BP mode.")

(defun bp-indent-line ()
  "Indent current line as BP code."
  (interactive)
  (beginning-of-line)
  (if (bobp)
      (indent-line-to 0)		         ; First line is always non-indented
      (let ((not-indented t) cur-indent)
        (cond ((looking-at "^[ \t]*end[ \t]*\\($\\|#.*$\\)")
               (save-excursion
                 (while not-indented
                   (forward-line -1)
                   (if (looking-at "^[ \t]*\\(blueprint\\|class\\)[ \t]+[_a-zA-Z0-9]+[ \t]*\\($\\|#.*$\\)")
                       (progn
                         (setq cur-indent (- (current-indentation) 0))
                         (setq not-indented nil))
                       (if (bobp)
                           (setq not-indented nil))  
                   ))))
              ((looking-at "^[ \t]*}.*$")
               ; set the indent to the previous line - 2
               (progn
                 (save-excursion
                   (forward-line -1)
                   (setq cur-indent (- (current-indentation) 2)))
                 (if (< cur-indent 0) ; We can't indent past the left margin
                     (setq cur-indent 0))))
              (t
               ; Iterate backwards until we find an indentation hint
               (save-excursion
                 (while not-indented
                   (forward-line -1)
                    ; [whitespace] blueprint/class/action name [# comment]
                    ; ... {
                   (if (looking-at "\\(^.*{[ \t]*\\($\\|#.*$\\)\\|^[ \t]*\\(blueprint\\|class\\|action\\)[ \t]+[_a-zA-Z0-9]+[ \t]*\\($\\|#.*$\\)\\)")
                       (progn
                         (setq cur-indent (+ (current-indentation) 2)) ; Do the actual indenting
                         (setq not-indented nil))
                       (if (looking-at "\\(^[ \t]*end\\|^.*}\\)") ; assume that the closing brace puts us at the same level
                           (progn
                             (setq cur-indent (+ (current-indentation) 0))
                             (setq not-indented nil))
                           (if (bobp)
                               (setq not-indented nil))
                       )
                   )
                 )
               ))
        )
        (if cur-indent
            (indent-line-to cur-indent)
            (indent-line-to 0))
))) ; If we didn't see an indentation hint, then allow no indentation

(defvar bp-mode-syntax-table nil
  "Syntax table for bp-mode.")

(defun bp-create-syntax-table ()
  (if bp-mode-syntax-table
	  ()
	(setq bp-mode-syntax-table (make-syntax-table))
	
    ; This is added so entity names with underscores can be more easily parsed
	(modify-syntax-entry ?_ "w" bp-mode-syntax-table)
  
	; line comments started with '#'
	(modify-syntax-entry ?#  "< b" bp-mode-syntax-table)
	(modify-syntax-entry ?\n "> b" bp-mode-syntax-table))
  (set-syntax-table bp-mode-syntax-table))

(defun bp-mode ()
  "Major mode for editing ORTS Blueprint files."
  (interactive)
  (kill-all-local-variables)
  (bp-create-syntax-table)
  
  ;; Set up font-lock
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults
		'(bp-font-lock-keywords))
  
  ;; Register our indentation function
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'bp-indent-line)
  
  (setq major-mode 'bp-mode)
  (setq mode-name "BP")
  (run-hooks 'bp-mode-hook))

(provide 'bp-mode)

;;; bp-mode.el ends here



