/** @page clientcode Client Code Description

- @ref clientcodeoverview
- @ref gamestateaccess
  - @ref gamestatemodulefunctions
  - @ref gamefunctions
- @ref pacecontrol
- @ref actionexecution
- @ref pathfinding
- @ref testingclientcode

@section clientcodeoverview Overview

The client code consists of a bunch of modules which communicate with each
other through events and function calls. The central one is GameStateModule
which connects to the server and maintains the state information.
Triggered by calling GameStateModule::recv_view() it will call registered
event handlers when a message from the server arrives. After setting actions
in objects, calling GameStateModule::send_actions() will send them to the
server (see below).

If you want to use the 3d gui (GfxModule) in your client have a look at
ortsg_main.C .
In a nutshell, you have to create a GfxModule object (say gfxm), hook it up
to the GameStateModule, and wire the event handlers.

For current code samples see orts_main.C and ortsg_main.C .


@section gamestateaccess Game State Access

@subsection gamestatemodulefunctions Game State Module Functions

GameStateModule::get_game()    - returns reference to game

GameStateModule::get_changes() - returns reference to structure that
contains lists of incremental game state changes (see GameChanges).
Now your event handler can act upon incremental changes, which may simplify
your code.

@subsection gamefunctions Game Functions

after
@code
Game &game = gsm.get_game();
@endcode
you can access the game state using the following functions:

- Game::get_map() - returns reference to map (basically an array of GameTiles)
- Game::get_boundaries() - vector of pointers to line segments defining
  terrain obstacles
- Game::get_objs() - returns vector of pointers to game objects owned by
  player index (index == Game::get_player_num() => neutral objects)
    
Examples of how to access the game state are given in GUI.C .

After setting actions of particular units using GameObj::set_action() you call
GameStateModule::send_actions() to send all actions to the server.

Accessing object properties is done like this:
@code
GameObj* gameobj;
gameobj->get_int("x"); 
@endcode
or 
@code
gameobj->sod.?  // ("server object data") 
@endcode
for internal properties like the owner of an object (see ServerObjData.H)
An important flag is gameobj->sod.in_game which indicates whether the object
is a game object (rather than a gfx object).

To access global player information such as resources, you do the following:

@code
PlayerInfo &pi = game.get_cplayer_info();
GameObj *player = pi.global_object_map["player"];
int my_gold = player->get_int("gold");
@endcode

Data members in sub-objects can be accessed like so:

@code
gameobj->components["armor"]->get_int("AC")
@endcode

See testgame/ for scripted game specifications (.bp files).


@section pacecontrol Pace Control

Class Game has four functions that can be called to obtain overall 
information about executed actions:

@code
sint4 Game::get_view_frame() const;
sint4 Game::get_action_frame() const;
sint4 Game::get_merged_actions() const;
sint4 Game::get_skipped_actions() const;
@endcode

<b>View frame</b>: refers to the game simulation step count. 

<b>%Action frame</b>: refers to the number of action messages (sent by the 
client) that were received before the current view was computed, 
and hence whose results are reflected in the current world view 
that the client has. Note that each action message includes up to 
one action per object.

It is assumed that the client sends an action message in
response to every view message it receives.

<b>Merged actions</b>: If due to network delays, or long processing between
calls to GameStateModule::send_actions(), multiple action messages arrive
late, they get merged into one action message to be executed next.

<b>Skipped actions</b>: %Action messages expire and their contents totally 
ignored if they lag too far behind the game simulation steps.

To learn about when actions get merged or skipped please have a look
at Server::recv_messages().


@section actionexecution Action Execution

There is also a mechanism that allows clients to know whether a 
specific action for a specific object
- <b>was skipped over</b> (due to delays, as mentioned); or
- <b>got to the server</b>, was executed and its results integrated in the 
  currently available view in the client; or
- <b>is pending</b>, meaning that it was not yet sent, or was already sent 
  but still did not complete a full 'trip' to the server and back, 
  and therefore the current view still reflects the world state 
  before the action was applied.

To get this info for a given GameObj, the client can:
- call GameObj::is_pending_action() which is true iff there is at least one
  pending action for this object.
- look into GameObj::pending_actions, which lists all actions that are
  still pending. Specific actions can be identified by the 
  client_action_id returned by the GameObj::set_action() call.
- use the incremental game changes that are reported with each
  view event by GameStateModule::get_changes(), listing all executed and
  skipped actions for all objects.


@section pathfinding Pathfinding

To get started with pathfinding have a look at libs/newpath/src
and at libs/serverclient/TerrainModule.*
There you will find pathfinding code that interacts with the 
ORTS client class and the graphical user interface. It's not
that sophisticated (and a bit buggy, too), but it can give 
you a sense on how to plan and execute paths in ORTS.


@section testingclientcode Testing Client Code Under Real-World Network Conditions

GameStateModule options -dmin &lt;int&gt; -dmax &lt;int&gt; and
-smin &lt;int&gt; -smax &lt;int&gt; can create randomized symmetric up/down
network lag (-d??? in millisecs) and put the client to sleep (-s???) for a
while when a view is received.
To see this in action issue:

@verbatim
orts @endverbatim 
and
@verbatim
orts -m client -act -dmin 500 -dmax 1000 -smin 100 -smax 300 @endverbatim
*/
