/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
  $header$
  $log$
*/

#ifndef RTS_Game_H
#define RTS_Game_H

#include "Global.H"
#include "Coord.H"
#include "String.H"
#include "Array.H"
#include <vector>
#include <list>
#include <SDL.h>
#include "RTS_Options.H"
#include "ID_Factory.H"
#include "Compress.H"
#include "Tile.H"

#define TEST_OUT 0

class Element;
class RT_GameParam;
class Action;
class Prop;
class BoolMatrix;

struct CollPair { sint4 ind1, ind2; sint8 coll_time; };

inline bool operator < (const CollPair &c1, const CollPair &c2) {
  return c1.coll_time < c2.coll_time;
}

typedef vector<CollPair> CollList;

typedef uint4 ObjId;
typedef map<uint4,uint4> IdMap;

class RTS_PlayerInfo {
    
public:
    
  string name;
  ID_Factory idf;
  IdMap sid2cid;  // server id -> client id
  IdMap cid2sid;  // client id -> server id
  set<ObjId> prev_obj_cids;  // c'ids of objs known from previous frame
  
  RTS_PlayerInfo();

  // delete cid,sid and frees cid
  void delete_cid(ObjId cid);
  
  // adds sid, new cid, return cid
  ObjId add_sid(ObjId sid);

  // 0 iff sid not found
  ObjId get_cid(ObjId sid);

  void check() const;

};


class RTS_Game {

public:

  enum { MAX_ELEM_NUM = 10000 }; // maximum number of objects
  enum { PLAYER_MAX_NUM = 16 };
  
protected:

  enum { UPD_CHAR = '0', NEW_CHAR='1', DEL_CHAR='2' };
  
  Pos dim;
 
  // TimeServer ts;
  // Scalar     time_step;  // mod 2 = 0

  const RTS_Options &opt;
  vector<Element*> elems;  // list of pointers to all game elements
  vector<Tile> tiles;
  vector<sint2> vis_counters[PLAYER_MAX_NUM];
  
  ID_Factory idf;
  IdMap id2index;          // ids -> indices in client mode
  
  SDL_Surface *screen;
  sint4 dim_x, dim_y;      // playing field dimensions, mesh_width
  sint4 pixel_scale;       // pixels/mesh_width
  sint4 s_x, s_y;          // screen dimensions
  sint4 sub_grid_res_bits; // ceil(log2(grid_width/sub_grid_width))
  sint4 sub_grid_res;      // 1<<sub_grid_res_bits

  sint4 visit_n;
  vector<RTS_PlayerInfo> players;

  BoolMatrix *x_shares_view_with_y, *prev_x_shares_view_with_y;
  BoolMatrix *x_is_friend_of_y, *prev_x_is_friend_of_y;

  typedef Array<uint4,50> SectorInfo;
  
  SectorInfo *sectors;   // for collision and visibility
  sint4 sector_num; 
  
  void refresh_views(); //Player *player, RTS_GameView *view, String &diff);

  real8 collision_time(Element &e1, Element &e2);
  real8 collision_time2(Element &e1, Element &e2, real8 ct);
  bool check_intersections(sint4 &ind1, sint4 &ind2);

  typedef list<sint4> Component;
  
  bool visit(const vector<Element*> &elems, sint4 i, sint4 *&comp);
  
  void handle_component(vector<Element*> &elems, sint4 *comp, sint4 n);

  void alloc_sectors(sint4 n);
  void free_sectors();
  void free_elems();  


  void elem_move_to_grid(Element *e);
  void elem_stop(Element *e);
  void elem_move_and_stop(Element *e, real8 t);
  void elem_move(Element *e, real8 t);
  bool elem_move_to(Element *e, Pos &tpos);

  inline Scalar fine_grid(sint4 x) { return x << sub_grid_res_bits; }
  void check_elements() const;
  
public:

  static const String MSG_COLLISION;
  FILE *hist_os[2];

  RTS_Game(const RTS_Options &opt);
  virtual ~RTS_Game();

  bool check_setup(ostream &os) const;
  bool init();
  // bool add_obj(const Prop &f, bool first_obj);
  bool compute_actions();
  void compute_visible_elements();  

  // call update_prev_prop() after creating all views!
  void create_view_diff(Compress &comp, sint4 player,
			vector<string> &view, uint4 &checksum);
  void update_view(Decompress &decomp, const vector<string> &obj_msgs);


  void create_action_messages(Compress &comp, sint4 cpind, vector<string> &msgs);
  void update_actions(Decompress &decomp, sint4 cpind, vector<string> &msgs);

  void create_vicinities(sint4 cpind, vector< vector<Prop> > &vics);

  void clear_actions();
  bool set_action(sint4 player, ObjId id, Action &act);

  void update_prev_prop();
  void refresh_id2index();
  
  void exec_actions();
  void remove_dead();
  void move_elements();
  bool finished(sint4 &winner);

  bool save(ostream &os) const;
  bool load(istream &is);

  void simulation();
  bool simulation_round(sint4 tick);
  void display(uint4 ind1 = uint4(-1), uint4 ind2 = uint4(-1));

  void prop_vectors(vector<Prop> &objs) const;
  
  static bool check_param(ostream &oserr, RT_GameParam &param);

  uint4 elems_checksum();
};

#endif

// boundaries of buildings, water, elevation
// computed from tiles and elements
//vector<LineSegments> lines;

