/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "RTS_Game.H"
#include "Element.H"
#include "SDL_gfxPrimitives.h"
#include "PairHeap.H"
#include "CycleCounter.H"
#include "Array.H"
#include <algorithm>
#include <SDL.h>

sint4 c1, c2;
uint8 total_ticks = 0;
uint8 total_n = 0;

uint8 t_total = 0;
uint8 t_sec   = 0;
uint8 t_edges = 0;
uint8 t_comp  = 0;
uint8 t_visit = 0;
uint8 t_mov   = 0;
uint8 t_heap  = 0;
uint8 t_n     = 0;

const bool INTERSECTION_CHECK = 0;

const sint4 SL = 3;

#define CHECK_INTERSECTIONS(x)				\
      if (INTERSECTION_CHECK) {							\
	sint4 a, b;					\
	if (check_intersections(a, b)) {		\
	  cout << "FOO! " << x << " " << a << " " << b << endl;	\
	  display(a, b); FOREVER{}; exit(1);					\
	}						\
      }

#define TOO_CLOSE(x, a, b)				\
      {							\
	cout << "TOO CLOSE " << x << " " << a << " " << b << endl;	\
	display(a, b); FOREVER{}; exit(1);					\
      }

void RTS_Game::check_elements() const
{
  uint4 n = elems.size();
  uint4 i;
  string s;
  
  for (i=0; i < n; i++) {

    const sint4 SUSP_RANGE  = 10000;
    const sint4 SUSP_DAMHIT = 1000;    
    Element &e = *elems[i];

    if (e.sv_id == 0 || e.sv_id >= 2000000000) { s = "id out of range"; break; }
    if (e.sv_owner < 1 || e.sv_owner > 16) { s = "owner out of range"; break; }
    if (e.sv_pos_x <= 0 || e.sv_pos_x >= dim_x) { s = "x out of range"; break; }
    if (e.sv_pos_y <= 0 || e.sv_pos_y >= dim_y) { s = "y out of range"; break; }
    if (e.sv_radius <= 0 || e.sv_radius >= SUSP_RANGE)  { s = "radius out of range"; break; }
    if (e.sv_sight_range <= 0 || e.sv_sight_range >= SUSP_RANGE) { s = "sigh range out of range"; break; }
    if (e.sv_min_attack_range >= SUSP_RANGE) { s = "min. attack range out of range"; break; }
    if (e.sv_max_attack_range >= SUSP_RANGE) { s = "max. attack range out of range"; break; }
    if (e.sv_speed >= SUSP_RANGE) { s = "speed out of range"; break; }
    if (e.sv_damage >= SUSP_DAMHIT) { s = "damage out of range"; break; }
    if (e.sv_hit_points >= sint4(SUSP_DAMHIT)) { s = "hit points out of range"; break; }
    if (e.sv_moving != 0 && e.sv_moving != 1) { s = "moving out of range"; break; }
  }

  if (i < n) {
    elems[i]->write();
    ERR(s);
  }
}


void RTS_Game::elem_move_to_grid(Element *e) {
  Pos g_pos = e->mv.fine_pos;
  g_pos.round(sub_grid_res_bits);
  e->sv_pos_x = g_pos.x >> sub_grid_res_bits;
  e->sv_pos_y = g_pos.y >> sub_grid_res_bits;
}

// stop object

void RTS_Game::elem_stop(Element *e) {
  if (e->sv_moving) {
    elem_move_to_grid(e);
    e->sv_moving = false;
  }
}

// advance position g_pos by t*heading_left
// set sv_pos to round(g_pos)

void RTS_Game::elem_move_and_stop(Element *e, real8 t) {
  elem_move(e, t);
  elem_stop(e);
}

// advance position g_pos by t*heading_left, adjust heading_left

void RTS_Game::elem_move(Element *e, real8 t) {
  Pos delta = e->sv_fine_heading;
  if (t != 1.0) delta.scale(t);
  e->mv.fine_pos.add(delta);
  
#if TEST    
  cout << "MOVED BY "; delta.write(); cout << endl;
#endif
}

bool RTS_Game::elem_move_to(Element *e, Pos &tpos) {

  if (tpos.bad()) {
    tpos.write(cerr);
    cerr << "target out of range" << endl;
    return true;
  }

  e->mv.target_pos = tpos;

  // sv_heading = delta*speed/||delta||

  Pos delta = e->mv.target_pos;
  Pos pos(e->sv_pos_x, e->sv_pos_y);
    
  // cout << "target= "; delta.write(); cout << endl;
  // cout << "curpos= "; sv_pos.write(); cout << endl;    
    
  delta.sub(pos);
  delta.x = fine_grid(delta.x);
  delta.y = fine_grid(delta.y);    
  Scalar dist = delta.l2_norm();
  if (dist == 0) return false;
    
  real8 f = real8(FixPoint::to_sint8(fine_grid(e->sv_speed)))/dist;
  delta.scale(f);
  e->sv_fine_heading = delta;

  //cout << "heading: "<< sv_heading.x << " " << sv_heading.y << endl;
    
  e->mv.fine_pos.x = fine_grid(e->sv_pos_x);
  e->mv.fine_pos.y = fine_grid(e->sv_pos_y);
    
  e->mv.ticks_so_far = 0;
  e->mv.ticks_total = sint4(ceil(1.0/f));

  if (e->mv.ticks_total <= 0) ERR("ticks < 0");

  e->sv_moving = true;
  return false;
}


void RTS_Game::move_elements()
{
  CycleCounter start_time;
  uint4 n = elems.size();

  if (n <= 0) return;
  
  // compute motion bounding circles

  Scalar fine_r_max = -1;
  
  FORU (i, n) {

    Element &e = *elems[i];

    if (e.sv_moving) {

      if (e.mv.ticks_total <= 0) cout << "FOO " << e.mv.ticks_total << endl;
      
      if (e.mv.ticks_so_far + 1 == e.mv.ticks_total) {

	// last tick => adjust heading such that target is reached with last step

	Pos delta = e.mv.target_pos;
	delta.x = fine_grid(delta.x);
	delta.y = fine_grid(delta.y);	
	delta.sub(e.mv.fine_pos);
	e.sv_fine_heading = delta;
      }
      
      // moving object: movement circle
      // (-x-)-->--M----(-x-)   
  
      // m = pos + (heading*0.5)

      Pos h = e.sv_fine_heading;
      h.scale(0.5);
      
      e.mv.bc_fine_pos = e.mv.fine_pos;
      e.mv.bc_fine_pos.add(h);
      
      e.mv.bc_fine_r = FixPoint::add(h.l2_norm(), fine_grid(e.sv_radius) + 2);
      // + 2: compensate for rounding error

    } else {

      // still object: bounding circle

      e.mv.bc_fine_pos.x = fine_grid(e.sv_pos_x);
      e.mv.bc_fine_pos.y = fine_grid(e.sv_pos_y);
      e.mv.bc_fine_r     = fine_grid(e.sv_radius);
    }

    fine_r_max = ::max(fine_r_max, e.mv.bc_fine_r);
  }

  // determine sector membership

  const sint4 K   = 8*fine_r_max/2;  // sector length > 2*fine_r_max
  const sint4 SX  = fine_grid(dim_x);
  const sint4 SY  = fine_grid(dim_y);
  const sint4 SNX = (SX+K-1)/K;
  const sint4 SNY = (SY+K-1)/K;
  const sint4 SNUM = SNX*SNY;
  uint1 active[SNUM];
  Array<sint4, 600> active_sectors;

  memset(active,  0, SNUM*sizeof(active[0]));

  // cout << SNX*SNY << " sectors " << endl;

  alloc_sectors(SNUM);

  FORS (i, SNUM) sectors[i].set_size(0);
  
  FORU (i, n) {

    Element &e = *elems[i];
    sint4 x = e.mv.bc_fine_pos.x;
    sint4 y = e.mv.bc_fine_pos.y;
    sint4 r = e.mv.bc_fine_r;
    Array<sint4,5> sl;
    sint4 dx[4] = { r, -r,  r, -r };
    sint4 dy[4] = { r,  r, -r, -r };    

    e.visited = false;
    e.neighbors.set_size(0);
    
    FORS (j, 4) {
      
      sint4 xi = x + dx[j];
      sint4 yi = y + dy[j];
      
      if (xi < 0) xi = 0; else if (xi >= SX) xi = SX-1;
      if (yi < 0) yi = 0; else if (yi >= SY) yi = SY-1;

      xi /= K;
      yi /= K;
      
      sint4 s = yi*SNX + xi;

      if (s >= SNUM) ERR("illegal sector");
      
      sint4 k;
      FOR (k, sl.size()) if (sl[k] == s) break;

      if (k >= sl.size()) {

	sl.push_back(s); // new sector member

	if (!active[s]) {
	  active[s] = true;
	  active_sectors.push_back(s);
	}
      }
    }

    FORS (j, sl.size()) {
      sectors[sl[j]].push_back(i); // update sector info
    }
  }

  CycleCounter end_time1;
  t_sec += end_time1.count() - start_time.count();

  
  // generate all pairs of elements that potentially collide

  c1 = c2 = 0;

  FORS (it, active_sectors.size()) { // all active sectors ...

    SectorInfo &si = sectors[active_sectors[it]];

    FORS (i1, si.size()) {

      sint4 n1 = si[i1];
      Element &e1 = *elems[n1];
      
      for (sint4 i2 = i1+1; i2 < si.size(); i2++) {

	// do movement circles intersect?

	sint4 n2 = si[i2];
	Element &e2 = *elems[n2];

	c1++;
	
	if (e1.sv_moving || e2.sv_moving) {

	  c2++;
	    
	  Scalar rsum = e1.mv.bc_fine_r + e2.mv.bc_fine_r + SL*sub_grid_res;
	  
	  if (e1.mv.bc_fine_pos.max_dist(e2.mv.bc_fine_pos) <= rsum) {

	    Scalar dist = e1.mv.bc_fine_pos.l2_dist(e2.mv.bc_fine_pos);

#if 0	
	    cout << n1 << " " << n2 << " " << dist  << " "
		 << (e1.mv.bc_r + e2.mv.bc_r + 2*sub_grid_res) << endl;
#endif
	    
	    if (dist <= rsum) {
	      
	      // potential collision

#if TEST_OUT	    
	      cout << "EDGE (" << n1 << "," << n2 << ")" << endl;
#endif

#if 0	      
	      cout << i1 << " " << i2 << endl;
	      cout << n1 << " " << n2 << endl;
	      cout << neighbors[n1].size() << " " << neighbors[n2].size() << endl;
#endif
	      e1.neighbors.push_back(n2);
	      e2.neighbors.push_back(n1);
	    }
	  }
	}
      }
    }
  }

  sint4 comp[n+1];

  // make neighbor lists unique (increases speed by 4%)
  // comp is used as a node set

  sint4 val = 1;
  
  memset(comp, 0, n*sizeof(comp[0]));

  if (n > 1000000000) ERR("n too large");
  
  FORU (i, n) {
    sint4 j = 0;
    NeighborInfo &a = elems[i]->neighbors;

    FORS (ni, a.size()) {
      sint4 ne = a[ni];
      if (comp[ne] != val) { a[j++] = ne; comp[ne] = val; }
    }

    a.set_size(j);
    val++;
  }

  CycleCounter end_time2;
  t_edges += end_time2.count() - end_time1.count();

  // handle movement in connected components

  visit_n = 0;

  FORU (i, n) {

    CycleCounter start_time;
    sint4 *p = comp;
    bool mobile = visit(elems, i, p);
    CycleCounter stop_time;
    t_visit += stop_time.count() - start_time.count();
    
    if (p != comp && mobile) {
      CycleCounter start_time;
      handle_component(elems, comp, p-comp);
      CycleCounter end_time;
      t_mov += end_time.count() - start_time.count();
    }
  }

  CycleCounter end_time3;

  bool stat;
  opt.get("-stat", stat);

  if (stat) {

    t_comp += end_time3.count() - end_time2.count();
    
    t_n++;
    t_total += end_time3.count() - start_time.count();

    cout << endl << "---------------" << endl;
    cout << "i=" << t_n << " n=" << n << endl;
    form(cout, "total:       %10lld\n", t_total/t_n);
    form(cout, "sectors:     %10lld %3lld%%\n", t_sec/t_n,   (t_sec*100LL)/t_total);
    form(cout, "edges:       %10lld %3lld%%\n", t_edges/t_n, (t_edges*100LL)/t_total);
    form(cout, "comp:        %10lld %3lld%%\n", t_comp/t_n,  (t_comp*100LL)/t_total);
    form(cout, "visit:       %10lld %3lld%%\n", t_visit/t_n, (t_visit*100LL)/t_total);
    form(cout, "movement:    %10lld %3lld%%\n", t_mov/t_n,   (t_mov*100LL)/t_total);
    form(cout, "heap:        %10lld %3lld%%\n", t_heap/t_n,  (t_heap*100LL)/t_total);

    form(cout, "alloc: %d %d\n", alloc_n/t_n, alloc_s/t_n);
    form(cout, "init:  %d\n", init_n/t_n);
  }

  CHECK_INTERSECTIONS("A");
}


// faster: update collision times in vicinity of colliding elements only

struct Edge {

public:

  sint4 i1, i2;    // i1 < i2

  Edge(sint4 i1_=-2, sint4 i2_=-1) {
    i1 = i1_; i2 = i2_;
    if (i1 == i2) ERR("LOOP");
    if (i1 > i2) { sint4 t = i1; i1 = i2; i2 = t; }
  }

  bool operator < (const Edge &e) const {
    if (i1 < e.i1) return true;
    if (i1 > e.i1) return false;
    return i2 < e.i2;
  }
};


#if 1

// sint4 seems to be sufficient

class CollTime {

public:

  sint4 v;

#define CT_FAC 100000000L  // 10^8
#define CT_FAC_INV (1.0L/CT_FAC)
  
  CollTime() { v = 0; }
  void from_real8(real8 f) { v = sint4(f * CT_FAC); }
  real8 to_real8() const { return real8(v) * CT_FAC_INV; }

  bool operator == (const CollTime &c) const { return v == c.v; }
  bool operator <  (const CollTime &c) const { return v < c.v; }  
  
};

#else

class CollTime {

public:

  sint8 v;

#define CT_FAC 1000000000000LL  // 10^12

  CollTime() { v = 0LL; }
  void from_real8(real8 f) { v = sint8(f * CT_FAC); }
  real8 to_real8() const { return real8(v)/CT_FAC; }

  bool operator == (const CollTime &c) const { return v == c.v; }
  bool operator <  (const CollTime &c) const { return v < c.v; }  
  
};

#endif


void RTS_Game::handle_component(vector<Element*> &elems, sint4 *comp, sint4 n)
{
  if (n <= 0) return;

#if TEST_OUT
  cout << "------------ COMPONENT ------------" << endl;

  { FORS (i, n) { cout << comp[i] << " "; } }
  cout << endl;
#endif

  PairHeap<Edge, CollTime> eh;
  sint4 stopped[n+1], *p_stopped, stopped_n;
  sint4 new_stopped[n+1], *p_new_stopped, new_stopped_n;
  
  // compute all collision times

  p_stopped = stopped;
  
  FORS (i, n) {

    sint4 i1 = comp[i];
    Element &e1 = *elems[i1];
    const NeighborInfo &ninf = e1.neighbors;
    
    FORS (ni, ninf.size()) {

      const sint4 ne = ninf[ni];

      if (ne < i1) {

	Element &e2 = *elems[ne];
	if (e1.sv_moving || e2.sv_moving) {

	  real8 ctf = collision_time2(e1, e2, 0.0);

#if TEST_OUT
	  cout << "CHECK " << i1 << " " << (ne) << endl;
	  cout << "CT " << ctf << endl;
#endif	    
	  if (ctf < 0.0) TOO_CLOSE("A", i1, ne);

	  if (ctf == 0.0) {

	    // immediate stop

	    if (e1.sv_moving) *p_stopped++ = i1;
	    if (e2.sv_moving) *p_stopped++ = ne;

	    elem_move_and_stop(&e1, 0.0);
	    elem_move_and_stop(elems[ne], 0.0);

#if TEST_OUT
	    cout << "STOPPED(0-1) " << i1 << " " << ne << endl;
#endif
	    
	  } else {

	    CollTime ct;
	    ct.from_real8(ctf);
	    Edge e(i1, ne);
	    eh.ins(e, ct);
	    
	  }
	}
      }
    }
  }

  CollTime curr_time;

  stopped_n = p_stopped - stopped;
  p_stopped = stopped;
  p_new_stopped = new_stopped;
  
  FOREVER {
  
    while (stopped_n > 0) {

      // update all collision times in the direct vicinity of stopped elements

      new_stopped_n = 0;
      
      FORS (i, stopped_n) {

	sint4 index = p_stopped[i];
	Element &e1 = *elems[index];
	NeighborInfo &ninfo = e1.neighbors;
	
	FORS (ni, ninfo.size()) {

	  const sint4 ne = ninfo[ni];
	  
	  Element &e2 = *elems[ne];	  
	  Edge e(index, ne);

	  if (e2.sv_moving) {

	    real8 ctf = curr_time.to_real8();
	    real8 diff = collision_time2(e1, e2, ctf);

	    if (diff < 0.0) TOO_CLOSE("A", p_stopped[i], ne);

	    if (diff == 0.0) {

	      // immediate stop
	      elem_move_and_stop(elems[ne], ctf);
	      p_new_stopped[new_stopped_n++] = ne;

#if TEST_OUT
	      cout << "STOPPED(0-2) " << p_stopped[i] << " " << ne << " " << ctf << endl;
#endif

	      // remove dead edge
	      eh.del(e);

	    } else if (ctf+diff >= 0.0 && ctf+diff <= 1.0L) {

	      // update collision time

	      CollTime ct;
	      ct.from_real8(ctf+diff);
	      eh.ins(e, ct);

#if TEST_OUT
	      cout << ">>>>>>>>>>>>>> UPDATE "
		   << e.i1 << " " << e.i2 << " " << ctf << " " << diff << endl;
#endif
	    } else {

	      // no collision anymore
	      CollTime ct;
	      ct.from_real8(2.0);
	      eh.ins(e, ct); 
	      
	    }
	    
	  } else {
	  
	    // remove dead edge
	    eh.del(e);
	  }
	}
      }

      sint4 *t = p_stopped;
      p_stopped = p_new_stopped;
      p_new_stopped = t;
      
      stopped_n = new_stopped_n;
    }

    Edge e;
    bool finished = false;
    
    FOREVER {
      if (!eh.min(e, curr_time) || curr_time.to_real8() >= 1.5) {
	finished = true; break;
      }
      if (elems[e.i1]->sv_moving || elems[e.i2]->sv_moving) break;
      eh.del(e);
    } 

    if (finished) break;

    real8 ctf = curr_time.to_real8();
    stopped_n = 0;
    
    if (elems[e.i1]->sv_moving) {
      elem_move_and_stop(elems[e.i1], ctf);
      stopped[stopped_n++] = e.i1;
    }
    if (elems[e.i2]->sv_moving) {
      elem_move_and_stop(elems[e.i2], ctf);
      stopped[stopped_n++] = e.i2;
    }

#if TEST_OUT
    cout << "STOPPED XXX " << e.i1 << " " << e.i2 << " " << ctf << endl;
#endif

    p_stopped = stopped;
    p_new_stopped = new_stopped;
  }


  // remaining moves

  {
    FORS (i, n) {

      real8 time_step = 1.0;
      Element &e = *elems[comp[i]];
    
      if (e.sv_moving) {
      
	elem_move(&e, time_step);
	elem_move_to_grid(&e);

#if TEST_OUT	
	cout << "NON_COLL MOVED " << comp[i];
	cout << " P="; e.mv.sg_pos.write();
	cout << endl;
#endif
	
	e.mv.ticks_so_far++;
	if (e.mv.ticks_so_far >= e.mv.ticks_total) {
	  e.sv_moving = false;
	}
      }
    }
  }

  t_heap += eh.ticks;
}
